# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
This module is to support vision visualization with opencv, which can help
developers use pre-trained models to predict and show the reasoning image fast.
Current it only supports object detection model.
"""
from . import object_detection
from .object_detection.object_detector import object_detection_predict, ObjectDetector
from .object_detection.utils.view_util import visualize_boxes_on_image, draw_boxes_on_image, save_image
from .object_detection.utils.config_util import load_and_parse_config


object_detection_utils = ['visualize_boxes_on_image', 'draw_boxes_on_image', 'save_image', 'load_and_parse_config']

__all__ = ['ObjectDetector', 'object_detection_predict']
__all__.extend(object_detection_utils)
__all__.extend(object_detection.__all__)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""SSD300 network based on MobileNetV2 backbone."""
import tinyms as ts
from .. import layers, primitives as P, Tensor
from ..vision.utils import ssd_default_boxes


def _make_divisible(v, divisor, min_value=None):
    """ensures that all layers have a channel number that is divisible by 8."""
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


def _conv2d(in_channel, out_channel, kernel_size=3, stride=1, pad_mod='same'):
    return layers.Conv2d(in_channel, out_channel, kernel_size=kernel_size, stride=stride,
                         padding=0, pad_mode=pad_mod, has_bias=True)


def _bn(channel):
    return layers.BatchNorm2d(channel, eps=1e-3, momentum=0.97,
                              gamma_init=1, beta_init=0, moving_mean_init=0, moving_var_init=1)


def _last_conv2d(in_channel, out_channel, kernel_size=3, stride=1, pad_mod='same', pad=0):
    in_channels = in_channel
    out_channels = in_channel
    depthwise_conv = layers.Conv2d(in_channels, out_channels, kernel_size, stride, pad_mode=pad_mod,
                                   padding=pad, group=in_channels)
    conv = _conv2d(in_channel, out_channel, kernel_size=1)
    return layers.SequentialLayer([depthwise_conv, _bn(in_channel), layers.ReLU6(), conv])


class ConvBNReLU(layers.Layer):
    """
    Convolution/Depthwise fused with Batchnorm and ReLU block definition.

    Args:
        in_channels (int): Input channel.
        out_channels (int): Output channel.
        kernel_size (int): Input kernel size. Default: 3.
        stride (int): Stride size for the first convolutional layer. Default: 1.
        groups (int): channel group. Convolution is 1 while Depthwise is input channel. Default: 1.

    Returns:
        Tensor, output tensor.

    Examples:
        >>> ConvBNReLU(16, 256, kernel_size=1, stride=1, groups=1)
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, groups=1):
        super(ConvBNReLU, self).__init__()
        padding = 0
        if groups == 1:
            conv = layers.Conv2d(in_channels, out_channels, kernel_size, stride,
                                 pad_mode='same', padding=padding)
        else:
            conv = layers.Conv2d(in_channels, in_channels, kernel_size, stride,
                                 pad_mode='same', padding=padding, group=in_channels)
        self.features = layers.SequentialLayer([conv,
                                                _bn(out_channels),
                                                layers.ReLU6()])

    def construct(self, x):
        output = self.features(x)
        return output


class InvertedResidual(layers.Layer):
    def __init__(self, inp, oup, stride, expand_ratio, use_relu=False):
        super(InvertedResidual, self).__init__()
        assert stride in [1, 2]

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = stride == 1 and inp == oup

        residual_layers = []
        if expand_ratio != 1:
            residual_layers.append(ConvBNReLU(inp, hidden_dim, kernel_size=1))
        residual_layers.extend([
            ConvBNReLU(hidden_dim, hidden_dim, stride=stride, groups=hidden_dim),
            layers.Conv2d(hidden_dim, oup, kernel_size=1, stride=1, has_bias=False),
            _bn(oup),
        ])
        self.conv = layers.SequentialLayer(residual_layers)
        self.use_relu = use_relu
        self.relu = layers.ReLU6()

    def construct(self, x):
        identity = x
        x = self.conv(x)
        if self.use_res_connect:
            return P.tensor_add(identity, x)
        if self.use_relu:
            x = self.relu(x)
        return x


class FlattenConcat(layers.Layer):
    """
    Concatenate predictions into a single tensor.

    Args:
        num_ssd_boxes (int): number of ssd boxes. Default is 1917.

    Returns:
        Tensor, flatten predictions.
    """

    def __init__(self, num_ssd_boxes=1917):
        super(FlattenConcat, self).__init__()
        self.num_ssd_boxes = num_ssd_boxes
        self.concat = P.Concat(axis=1)
        self.transpose = P.Transpose()

    def construct(self, inputs):
        output = ()
        batch_size = P.shape(inputs[0])[0]
        for x in inputs:
            x = self.transpose(x, (0, 2, 3, 1))
            output += (P.reshape(x, (batch_size, -1)),)
        res = self.concat(output)
        return P.reshape(res, (batch_size, self.num_ssd_boxes, -1))


class MultiBox(layers.Layer):
    """
    Multibox conv layers. Each multibox layer contains class conf scores
    and localization predictions.

    Args:
        class_num (int): number of classes. Default: 21.

    Returns:
        Tensor, localization predictions.
        Tensor, class conf scores.
    """

    def __init__(self, class_num=21):
        super(MultiBox, self).__init__()
        out_channels = [576, 1280, 512, 256, 256, 128]
        num_default = [3, 6, 6, 6, 6, 6]

        loc_layers = []
        cls_layers = []
        for k, out_channel in enumerate(out_channels):
            loc_layers += [_last_conv2d(out_channel, 4 * num_default[k],
                                        kernel_size=3, stride=1, pad_mod='same', pad=0)]
            cls_layers += [_last_conv2d(out_channel, class_num * num_default[k],
                                        kernel_size=3, stride=1, pad_mod='same', pad=0)]

        self.multi_loc_layers = layers.LayerList(loc_layers)
        self.multi_cls_layers = layers.LayerList(cls_layers)
        self.flatten_concat = FlattenConcat()

    def construct(self, inputs):
        loc_outputs = ()
        cls_outputs = ()
        for i in range(len(self.multi_loc_layers)):
            loc_outputs += (self.multi_loc_layers[i](inputs[i]),)
            cls_outputs += (self.multi_cls_layers[i](inputs[i]),)
        return self.flatten_concat(loc_outputs), self.flatten_concat(cls_outputs)


class SSDWithMobileNetV2(layers.Layer):
    """
    MobileNetV2 backbone.

    Args:
        width_mult (float): Channels multiplier for round to 8/16 and others. Default: 1.0.
        round_nearest (int): Channel round to. Default: 8.
        input_channel (int): Input channel. Default: 32.
        last_channel (int): The channel of last layer. Default: 1280.

    Returns:
        Tensor, output feature maps.
        Tensor, output tensor.
    """

    def __init__(self, width_mult=1.0, round_nearest=8, input_channel=32, last_channel=1280):
        super(SSDWithMobileNetV2, self).__init__()
        # setting of inverted residual blocks
        cfg = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]

        # building first layer
        input_channel = _make_divisible(input_channel * width_mult, round_nearest)
        self.last_channel = _make_divisible(last_channel * max(1.0, width_mult), round_nearest)
        backbone_layers = [ConvBNReLU(3, input_channel, stride=2)]
        # building inverted residual blocks
        layer_index = 0
        for t, c, n, s in cfg:
            output_channel = _make_divisible(c * width_mult, round_nearest)
            for i in range(n):
                if layer_index == 13:
                    hidden_dim = int(round(input_channel * t))
                    self.expand_layer_conv_13 = ConvBNReLU(input_channel, hidden_dim, kernel_size=1)
                stride = s if i == 0 else 1
                backbone_layers.append(InvertedResidual(input_channel, output_channel, stride, expand_ratio=t))
                input_channel = output_channel
                layer_index += 1
        # building last several layers
        backbone_layers.append(ConvBNReLU(input_channel, self.last_channel, kernel_size=1))

        self.features_1 = layers.SequentialLayer(backbone_layers[:14])
        self.features_2 = layers.SequentialLayer(backbone_layers[14:])

    def construct(self, x):
        out = self.features_1(x)
        expand_layer_conv_13 = self.expand_layer_conv_13(out)
        out = self.features_2(out)
        return expand_layer_conv_13, out


class SSD300(layers.Layer):
    """
    SSD300 Network. Default backbone is MobileNetV2.

    Args:
        backbone (layers.Layer): backbone of ssd300 model.
        class_num (int): number of classes. Default: 21.
        is_training (bool): Specify if in training step. Default: True.

    Returns:
        Tensor, localization predictions.
        Tensor, class conf scores.
    """

    def __init__(self, backbone, class_num=21, is_training=True):
        super(SSD300, self).__init__()

        self.backbone = backbone
        in_channels = [256, 576, 1280, 512, 256, 256]
        out_channels = [576, 1280, 512, 256, 256, 128]
        ratios = [0.2, 0.2, 0.2, 0.25, 0.5, 0.25]
        strides = [1, 1, 2, 2, 2, 2]
        residual_list = []
        for i in range(2, len(in_channels)):
            residual = InvertedResidual(in_channels[i], out_channels[i], stride=strides[i],
                                        expand_ratio=ratios[i], use_relu=True)
            residual_list.append(residual)
        self.multi_residual = layers.LayerList(residual_list)
        self.multi_box = MultiBox(class_num=class_num)
        self.is_training = is_training
        self.activation = P.Sigmoid()

    def construct(self, x):
        layer_out_13, output = self.backbone(x)
        multi_feature = (layer_out_13, output)
        feature = output
        for residual in self.multi_residual:
            feature = residual(feature)
            multi_feature += (feature,)
        pred_loc, pred_label = self.multi_box(multi_feature)
        if not self.is_training:
            pred_label = self.activation(pred_label)
        pred_loc = P.cast(pred_loc, ts.float32)
        pred_label = P.cast(pred_label, ts.float32)
        return pred_loc, pred_label


class SSDInferWithDecoder(layers.Layer):
    """
    SSD300 infer wrapper to decode the bbox locations.

    Args:
        network (layers.Layer): the origin ssd300 infer network without bbox decoder.

    Returns:
        Tensor, the locations for bbox after decoder representing (y0, x0, y1, x1)
        Tensor, the prediction labels.
    """

    def __init__(self, network):
        super(SSDInferWithDecoder, self).__init__()
        self.network = network
        self.default_boxes = Tensor(ssd_default_boxes, dtype=ts.float32)
        prior_scaling_tensor = Tensor((0.1, 0.2), dtype=ts.float32)
        self.prior_scaling_xy = prior_scaling_tensor[0]
        self.prior_scaling_wh = prior_scaling_tensor[1]

    def construct(self, x):
        pred_loc, pred_label = self.network(x)

        default_bbox_xy = self.default_boxes[..., :2]
        default_bbox_wh = self.default_boxes[..., 2:]
        pred_xy = pred_loc[..., :2] * self.prior_scaling_xy * default_bbox_wh + default_bbox_xy
        pred_wh = P.Exp()(pred_loc[..., 2:] * self.prior_scaling_wh) * default_bbox_wh

        pred_xy_0 = pred_xy - pred_wh / 2.0
        pred_xy_1 = pred_xy + pred_wh / 2.0
        pred_xy = P.Concat(-1)((pred_xy_0, pred_xy_1))
        pred_xy = P.Maximum()(pred_xy, 0)
        pred_xy = P.Minimum()(pred_xy, 1)
        return pred_xy, pred_label


def ssd300_mobilenetv2(**kwargs):
    """
    Get SSD300 model instance for training, evaluation and prediction.

    Args:
        class_num (int): The number of classes. Default: 21.
        is_training (bool): Whether to do training job, default: True.

    Returns:
        model.SSD300, SSD300 instance.
    """
    net = SSD300(SSDWithMobileNetV2(), class_num=kwargs.get('class_num', 21),
                 is_training=kwargs.get('is_training', True))
    if kwargs.get('is_training') is False:
        net = SSDInferWithDecoder(net)

    return net



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np

import tinyms as ts
from tinyms import layers, Tensor
from tinyms.primitives import tensor_add, Softmax, ReduceMean


def _make_divisible(v, divisor, min_value=None):
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class GlobalAvgPooling(layers.Layer):
    """
    Pooling Layer, do average pooling operation.

    Returns:
        Tensor, output tensor.
    """

    def __init__(self):
        super(GlobalAvgPooling, self).__init__()
        self.mean = ReduceMean(keep_dims=False)

    def construct(self, x):
        x = self.mean(x, (2, 3))
        return x


class ConvBNReLU(layers.Layer):
    """
    Convolution/Depthwise fused with Batchnorm and ReLU block definition.

    Args:
        in_channels (int): Input channel.
        out_channels (int): Output channel.
        kernel_size (int): Input kernel size. Default: 3.
        stride (int): Stride size for the first convolutional layer. Default: 1.
        groups (int): channel group. Convolution is 1 while Depthwise is input channel. Default: 1.

    Returns:
        Tensor, output tensor.

    Examples:
        >>> ConvBNReLU(16, 256, kernel_size=3, stride=1, groups=1)
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, groups=1):
        super(ConvBNReLU, self).__init__()
        padding = (kernel_size - 1) // 2
        if groups == 1:
            conv = layers.Conv2d(in_channels, out_channels, kernel_size, stride,
                                 pad_mode='pad', padding=padding)
        else:
            conv = layers.Conv2d(in_channels, in_channels, kernel_size, stride,
                                 pad_mode='pad', padding=padding, group=in_channels)
        self.features = layers.SequentialLayer([conv,
                                                layers.BatchNorm2d(out_channels),
                                                layers.ReLU6()])

    def construct(self, x):
        output = self.features(x)
        return output


class InvertedResidual(layers.Layer):
    """
    Inverted residual module.

    Args:
        inp (int): Input channel.
        oup (int): Output channel.
        stride (int): Stride size.
        expand_ratio (float): Expand ratio.

    Returns:
        Tensor, output tensor.
    """

    def __init__(self, inp, oup, stride, expand_ratio):
        super(InvertedResidual, self).__init__()
        assert stride in [1, 2]

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = stride == 1 and inp == oup

        residual_layers = []
        if expand_ratio != 1:
            residual_layers.append(ConvBNReLU(inp, hidden_dim, kernel_size=1))
        residual_layers.extend([
            ConvBNReLU(hidden_dim, hidden_dim, stride=stride, groups=hidden_dim),
            layers.Conv2d(hidden_dim, oup, kernel_size=1, stride=1, has_bias=False),
            layers.BatchNorm2d(oup),
        ])
        self.conv = layers.SequentialLayer(residual_layers)

    def construct(self, x):
        identity = x
        x = self.conv(x)
        if self.use_res_connect:
            return tensor_add(identity, x)
        return x


class MobileNetV2Backbone(layers.Layer):
    """
    MobileNetV2 Backbone net.

    Args:
        width_mult (int): The number of classes.
        round_nearest (int): Channel round to. Default is 8.
        input_channel (int): Input channel. Default is 32.
        last_channel (int): The channel of last layer. Default is 1280.

    Returns:
        Tensor, output tensor.
    """

    def __init__(self, width_mult=1., round_nearest=8, input_channel=32, last_channel=1280):
        super(MobileNetV2Backbone, self).__init__()
        # setting of inverted residual blocks
        self.cfgs = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]

        # building first layer
        input_channel = _make_divisible(input_channel * width_mult, round_nearest)
        self.last_channel = _make_divisible(last_channel * max(1.0, width_mult), round_nearest)
        backbone_layers = [ConvBNReLU(3, input_channel, stride=2)]
        # building inverted residual blocks
        for t, c, n, s in self.cfgs:
            output_channel = _make_divisible(c * width_mult, round_nearest)
            for i in range(n):
                stride = s if i == 0 else 1
                backbone_layers.append(InvertedResidual(input_channel, output_channel, stride, expand_ratio=t))
                input_channel = output_channel
        # building last several layers
        backbone_layers.append(ConvBNReLU(input_channel, self.last_channel, kernel_size=1))
        self.backbone = layers.SequentialLayer(backbone_layers)
        self._initialize_weights()

    def construct(self, x):
        x = self.backbone(x)
        return x

    def _initialize_weights(self):
        self.init_parameters_data()
        for _, m in self.cells_and_names():
            if isinstance(m, layers.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.set_data(Tensor(np.random.normal(0, np.sqrt(2. / n),
                                                          m.weight.data.shape).astype("float32")))
                if m.bias is not None:
                    m.bias.set_data(ts.zeros(m.bias.data.shape))
            elif isinstance(m, layers.BatchNorm2d):
                m.gamma.set_data(ts.ones(m.gamma.data.shape))
                m.beta.set_data(ts.zeros(m.beta.data.shape))


class MobileNetV2Head(layers.Layer):
    """
    MobileNetV2 Head net.

    Args:
        input_channel (int): Input channel. Default is 1280.
        class_num (int): The number of classes. Default is 1000.
        use_activation (bool): Whether to use activation functions. Default is False.

    Returns:
        Tensor, output tensor.
    """

    def __init__(self, input_channel=1280, class_num=1000, use_activation=False):
        super(MobileNetV2Head, self).__init__()
        # mobilenet head
        self.head = layers.SequentialLayer(([GlobalAvgPooling(),
                                             layers.Dense(input_channel, class_num)]))
        self.use_activation = use_activation
        self.activation = Softmax()
        self._initialize_weights()

    def construct(self, x):
        x = self.head(x)
        if self.use_activation:
            x = self.activation(x)
        return x

    def _initialize_weights(self):
        self.init_parameters_data()
        for _, m in self.cells_and_names():
            if isinstance(m, layers.Dense):
                m.weight.set_data(Tensor(np.random.normal(
                    0, 0.01, m.weight.data.shape).astype("float32")))
                if m.bias is not None:
                    m.bias.set_data(ts.zeros(m.bias.data.shape))


class MobileNetV2(layers.Layer):
    """
    MobileNetV2 architecture.

    Args:
        class_num (int): The number of classes.
        width_mult (float): Channels multiplier for round to 8/16 and others. Default is 1.0.
        round_nearest (int): Channel round to. Default is 8.
        input_channel (int): Input channel. Default is 32.
        last_channel (int): The channel of last layer. Default is 1280.

    Returns:
        Tensor, output tensor.
    """

    def __init__(self, class_num=1000, width_mult=1.,
                 round_nearest=8, input_channel=32, last_channel=1280, is_training=True):
        super(MobileNetV2, self).__init__()
        self.backbone = MobileNetV2Backbone(width_mult=width_mult,
                                            round_nearest=round_nearest,
                                            input_channel=input_channel,
                                            last_channel=last_channel)
        self.head = MobileNetV2Head(input_channel=self.backbone.last_channel,
                                    class_num=class_num, use_activation=not is_training)

    def construct(self, x):
        x = self.backbone(x)
        x = self.head(x)
        return x


def mobilenetv2(**kwargs):
    """
    Get MobileNetV2 instance for model training, evaluation and prediction.

    Args:
        class_num (int): The number of classes. Default: 1000.
        is_training (bool): Whether to do training job, default: True.

    Returns:
        model.MobileNetV2, MobileNetV2 instance.
    """
    return MobileNetV2(class_num=kwargs.get('class_num', 1000),
                       is_training=kwargs.get('is_training', True))



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from easydict import EasyDict as ed

from tinyms import model

MODEL_HUB = ed({
    "alexnet_v1": model.alexnet,
    "lenet5_v1": model.lenet5,
    "resnet50_v1": model.resnet50,
    "mobilenet_v2": model.mobilenetv2,
    "ssd300_v1": model.ssd300_mobilenetv2,
    "vgg16_v1": model.vgg16,
})



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Layer module contains pre-defined building blocks or computing units to construct neural networks.

The high-level components (Layers) used to construct the neural network.
"""

from mindspore.nn import Cell, GraphCell
from mindspore.nn.layer import SequentialCell, CellList
from mindspore.nn import wrap
from mindspore.nn.wrap import *
from mindspore.nn import layer
from mindspore.nn.layer import *


__all__ = ['Layer', 'SequentialLayer', 'LayerList']
__all__.extend(wrap.__all__)
__all__.extend(layer.__all__)


class Layer(Cell):
    """
    Base class for all neural networks.

    A 'Layer' could be a single neural network layer, such as conv2d, relu, batch_norm, etc. or a composition of cells to constructing a network.

    Note:
        In general, the autograd algorithm will automatically generate the implementation of the gradient function, but if back-propagation(bprop) method is implemented, the gradient function will be replaced by the bprop.
        The bprop implementation will receive a Tensor `dout` containing the gradient of the loss w.r.t. the output, and a Tensor `out` containing the forward result.
        The bprop needs to compute the gradient of the loss w.r.t. the inputs, gradient of the loss w.r.t. Parameter variables are not supported currently.
        The bprop method must contain the self parameter.

    Args:
        auto_prefix (bool): Recursively generate namespaces. Default: True.

    Examples:
        >>> from tinyms import layers, primitives as P
        >>>
        >>> class MyNet(layers.Layer):
        ...    def __init__(self):
        ...        super(MyNet, self).__init__()
        ...        self.relu = P.ReLU()
        ...
        ...    def construct(self, x):
        ...        return self.relu(x)
    """
    pass


class GraphLayer(GraphCell):
    """
    Base class for running the graph loaded from MindIR.

    This feature is still under development. Currently `GraphLayer` do not support modifying the structure of the
    diagram, and can only use data that shape and type are the same as the input when exporting the MindIR.

    Args:
        graph (object): A compiled graph loaded from MindIR.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``
    """
    pass


class SequentialLayer(SequentialCell):
    """
    Sequential layer container.

    A list of Layers will be added to it in the order they are passed in the constructor.
    Alternatively, an ordered dict of cells can also be passed in.

    Args:
        args (Union[list, OrderedDict]): List of subclass of Layer.

    Raises:
        TypeError: If the type of the argument is not list or OrderedDict.

    Inputs:
        - **input** (Tensor) - Tensor with shape according to the first Cell in the sequence.

    Outputs:
        Tensor, the output Tensor with shape depending on the input and defined sequence of Layers.

    Examples:
        >>> import tinyms as ts
        >>> from tinyms.layers import SequentialLayer, Conv2d, ReLU
        >>>
        >>> seq_layer = SequentialLayer([Conv2d(3, 2, 3, pad_mode='valid', weight_init="ones"), ReLU()])
        >>> x = ts.ones([1, 3, 4, 4])
        >>> print(seq_layer(x))
        [[[[27. 27.]
           [27. 27.]]
          [[27. 27.]
           [27. 27.]]]]
    """
    pass


class LayerList(CellList):
    """
    Holds Layers in a list.

    LayerList can be used like a regular Python list, support
    '__getitem__', '__setitem__', '__delitem__', '__len__', '__iter__' and '__iadd__',
    but layers it contains are properly registered, and will be visible by all Layer methods.

    Args:
        args (list, optional): List of subclass of Layer.

    Examples:
        >>> from tinyms.layers import LayerList, Conv2d, BatchNorm2d, ReLU
        >>>
        >>> conv = nn.Conv2d(100, 20, 3)
        >>> layers = LayerList([BatchNorm2d(20)])
        >>> layers.insert(0, Conv2d(100, 20, 3))
        >>> layers.append(ReLU())
        >>> layers
        LayerList<
          (0): Conv2d<input_channels=100, ..., bias_init=None>
          (1): BatchNorm2d<num_features=20, ..., moving_variance=Parameter (name=variance)>
          (2): ReLU<>
          >
    """
    pass



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
import numpy as np
from PIL import Image
import tinyms as ts

from . import _transform_ops
from ._transform_ops import *
from .utils import ssd_bboxes_encode, ssd_bboxes_filter, jaccard_numpy
from .. import Tensor
from ..data import MnistDataset, Cifar10Dataset, ImageFolderDataset, VOCDataset, GeneratorDataset
from ..primitives import Softmax
from .transform_config import get_specified_config

__all__ = [
    'mnist_transform', 'MnistTransform',
    'cifar10_transform', 'Cifar10Transform',
    'imagefolder_transform', 'ImageFolderTransform',
    'voc_transform', 'VOCTransform',
    'shanshui_tranform', 'ShanshuiTransform',
    'cyclegan_transform', 'CycleGanDatasetTransform'
]
__all__.extend(_transform_ops.__all__)


class DatasetTransform(object):
    r'''
    Base class for all dataset transforms.
    '''

    def __init__(self, configs=None):
        if configs:
            self.configs = configs
        else:
            self.configs = get_specified_config('DatasetTransform')
        self.labels = self.configs['labels']
        self.transform_strategy = self.configs['transform_strategy']

    def apply_ds(self, ds, trans_func=None, repeat_size=1, batch_size=32,
                 num_parallel_workers=None):
        if not isinstance(trans_func, list):
            raise TypeError('trans_func must be list')

        # apply map operations on datasets
        ds = ds.map(operations=TypeCast(ts.int32), input_columns="label",
                    num_parallel_workers=num_parallel_workers)
        ds = ds.map(operations=trans_func, input_columns="image", num_parallel_workers=num_parallel_workers)
        # apply batch operations
        ds = ds.batch(batch_size, drop_remainder=True)
        # apply repeat operations
        ds = ds.repeat(repeat_size)

        return ds

    def postprocess(self, input, strategy='TOP1_CLASS'):
        r'''
        Apply postprocess operation for prediction result.

        Args:
            input (numpy.ndarray): Prediction result.
            strategy (str): Specifies the postprocess strategy. Default: TOP1_CLASS.

        Returns:
            str, the postprocess result.
        '''
        if not isinstance(input, np.ndarray):
            raise TypeError("Input should be NumPy, got {}.".format(type(input)))
        if not input.ndim == 2:
            raise TypeError("Input should be 2-D Numpy, got {}.".format(input.ndim))
        if strategy not in self.transform_strategy:
            raise ValueError("Strategy should be one of {}, got {}.".format(self.transform_strategy, strategy))

        softmax = Softmax()
        score_list = softmax(Tensor(input, dtype=ts.float32)).asnumpy()
        if strategy == 'TOP1_CLASS':
            score = max(score_list[0])
            return ('TOP1: {}, score: {}').format(str(self.labels[input[0].argmax()]), str(round(score, 5)))
        else:
            label_index = np.argsort(input[0])[::-1]
            score_index = np.sort(score_list[0])[::-1]
            top5_labels = []
            res = ''
            top5_scores = score_index[:5].tolist()
            top_num = int(strategy.split('_')[0].split('TOP')[-1])
            for i in range(top_num):
                top5_labels.append(self.labels[label_index[i]])
                res += 'TOP' + str(i+1) + ": " + str(top5_labels[i]) + \
                       ", score: " + str(format(top5_scores[i], '.7f')) + '\t'
            return res


class MnistTransform(DatasetTransform):
    r'''
    Mnist dataset transform class.

    Inputs:
        img (Union[numpy.ndarray, PIL.Image]): Image to be transformed in Mnist-style.

    Outputs:
        numpy.ndarray, transformed image.

    Examples:
        >>> from PIL import Image
        >>> from tinyms.vision import MnistTransform
        >>>
        >>> mnist_transform = MnistTransform()
        >>> img = Image.open('object_detection.jpg')
        >>> img = mnist_transform(img)
    '''

    def __init__(self, configs=None):
        if configs:
            self.configs = configs
        else:
            self.configs = get_specified_config('MnistTransform')
        super().__init__(configs=self.configs)
        self.labels = self.configs['labels']
        self.transform_strategy = self.configs['transform_strategy']
        self.grayscale = Grayscale()
        self.resize = Resize(self.configs['resize'])
        self.normalize = Rescale(eval(self.configs['rescale1']['rescale_factor']),
                                 eval(self.configs['rescale1']['shift_factor']))
        self.rescale = Rescale(eval(self.configs['rescale2']['rescale_factor']),
                               self.configs['rescale2']['shift_factor'])

    def __call__(self, img):
        if not isinstance(img, (np.ndarray, Image.Image)):
            raise TypeError("Input type should be numpy.ndarray or PIL.Image, got {}.".format(type(img)))
        if isinstance(img, np.ndarray):
            img = Image.fromarray(img.astype('uint8')).convert('RGB')
        img = np.asarray(self.grayscale(img), dtype=np.float32)
        img = np.expand_dims(img, 2)
        img = self.resize(img)
        img = self.normalize(img)
        img = self.rescale(img)
        img = hwc2chw(img)

        return img

    def apply_ds(self, mnist_ds, repeat_size=1, batch_size=32, num_parallel_workers=None):
        r'''
        Apply preprocess operation on MnistDataset instance.

        Args:
            mnist_ds (data.MnistDataset): MnistDataset instance.
            repeat_size (int): The repeat size of dataset. Default: 1.
            batch_size (int): Batch size. Default: 32.
            num_parallel_workers (int): The number of concurrent workers. Default: None.

        Returns:
            data.MnistDataset, the preprocessed MnistDataset instance.

        Examples:
            >>> from tinyms.vision import MnistTransform
            >>>
            >>> mnist_transform = MnistTransform()
            >>> mnist_ds = mnist_transform.apply_ds(mnist_ds)
        '''
        if not isinstance(mnist_ds, MnistDataset):
            raise TypeError("Input type should be MnistDataset, got {}.".format(type(mnist_ds)))

        trans_func = [self.resize, self.normalize, self.rescale, hwc2chw]
        # apply transform functions on mnist dataset
        mnist_ds = super().apply_ds(mnist_ds, trans_func=trans_func, repeat_size=repeat_size,
                                    batch_size=batch_size, num_parallel_workers=num_parallel_workers)

        return mnist_ds


class Cifar10Transform(DatasetTransform):
    r'''
    Cifar10 dataset transform class.

    Inputs:
        img (Union[numpy.ndarray, PIL.Image]): Image to be transformed in Cifar10-style.

    Outputs:
        numpy.ndarray, Transformed image.

    Examples:
        >>> from PIL import Image
        >>> from tinyms.vision import Cifar10Transform
        >>>
        >>> cifar10_transform = Cifar10Transform()
        >>> img = Image.open('object_detection.jpg')
        >>> img = cifar10_transform(img)
    """
    '''

    def __init__(self, configs=None):
        if configs:
            self.configs = configs
        else:
            self.configs = get_specified_config('Cifar10Transform')
        super().__init__(configs=self.configs)
        self.random_crop = RandomCrop(self.configs['random_crop']['size'], self.configs['random_crop']['padding'])
        self.random_horizontal_flip = RandomHorizontalFlip(prob=self.configs['random_horizontal_flip']['prob'])
        self.resize = Resize(self.configs['resize'])
        self.rescale = Rescale(eval(self.configs['rescale']['rescale_factor']),
                               self.configs['rescale']['shift_factor'])
        self.normalize = Normalize(self.configs['normalize']['mean'], self.configs['normalize']['std'])

    def __call__(self, img):
        if not isinstance(img, (np.ndarray, Image.Image)):
            raise TypeError("Input type should be numpy.ndarray or PIL.Image, got {}.".format(type(img)))
        img = self.resize(img)
        img = self.rescale(img)
        img = self.normalize(img)
        img = hwc2chw(img)

        return img

    def apply_ds(self, cifar10_ds, repeat_size=1, batch_size=32,
                 num_parallel_workers=None, is_training=True):
        r'''
        Apply preprocess operation on Cifar10Dataset instance.

        Args:
            cifar10_ds (data.Cifar10Dataset): Cifar10Dataset instance.
            repeat_size (int): The repeat size of dataset. Default: 1.
            batch_size (int): Batch size. Default: 32.
            num_parallel_workers (int): The number of concurrent workers. Default: None.
            is_training (bool): Specifies if is in training step. Default: True.

        Returns:
            data.Cifar10Dataset, the preprocessed Cifar10Dataset instance.

        Examples:
            >>> from tinyms.vision import Cifar10Transform
            >>>
            >>> cifar10_transform = Cifar10Transform()
            >>> cifar10_ds = cifar10_transform.apply_ds(cifar10_ds)
        '''
        if not isinstance(cifar10_ds, Cifar10Dataset):
            raise TypeError("Input type should be Cifar10Dataset, got {}.".format(type(cifar10_ds)))

        trans_func = []
        if is_training:
            trans_func += [self.random_crop, self.random_horizontal_flip]
        trans_func += [self.resize, self.rescale, self.normalize, hwc2chw]
        # apply transform functions on cifar10 dataset
        cifar10_ds = super().apply_ds(cifar10_ds, trans_func=trans_func, repeat_size=repeat_size,
                                      batch_size=batch_size, num_parallel_workers=num_parallel_workers)

        return cifar10_ds


class ImageFolderTransform(DatasetTransform):
    r'''
    ImageFolder dataset transform class.

    Inputs:
        img (Union[numpy.ndarray, PIL.Image]): Image to be transformed in ImageFolder-style.

    Outputs:
        numpy.ndarray, transformed image.

    Examples:
        >>> from PIL import Image
        >>> from tinyms.vision import ImageFolderTransform
        >>>
        >>> imagefolder_transform = ImageFolderTransform()
        >>> img = Image.open('object_detection.jpg')
        >>> img = imagefolder_transform(img)
    '''

    def __init__(self, configs=None):
        if configs:
            self.configs = configs
        else:
            self.configs = get_specified_config('ImageFolderTransform')
        super().__init__(configs=self.configs)
        self.random_crop_decode_resize = RandomCropDecodeResize(
            self.configs['random_crop_decode_resize']['size'],
            scale=tuple(self.configs['random_crop_decode_resize']['scale']),
            ratio=tuple(self.configs['random_crop_decode_resize']['ratio']))
        self.random_horizontal_flip = RandomHorizontalFlip(prob=self.configs['random_horizontal_flip']['prob'])
        self.resize = Resize(self.configs['resize'])
        self.center_crop = CenterCrop(self.configs['center_crop'])
        self.normalize = Normalize(self.configs['normalize']['mean'], self.configs['normalize']['std'])

    def _center_crop(self, img):
        y, x, _ = img.shape
        startx = x // 2 - (224 // 2)
        starty = y // 2 - (224 // 2)
        return img[starty:starty + 224, startx:startx + 224, :]

    def __call__(self, img):
        if not isinstance(img, (np.ndarray, Image.Image)):
            raise TypeError("Input type should be numpy.ndarray or PIL.Image, got {}.".format(type(img)))
        img = self.resize(img)
        img = self._center_crop(img)
        img = self.normalize(img)
        img = hwc2chw(img)

        return img

    def apply_ds(self, imagefolder_ds, repeat_size=1, batch_size=32,
                 num_parallel_workers=None, is_training=True):
        r'''
        Apply preprocess operation on ImageFolderDataset instance.

        Args:
            cifar10_ds (data.ImageFolderDataset): ImageFolderDataset instance.
            repeat_size (int): The repeat size of dataset. Default: 1.
            batch_size (int): Batch size. Default: 32.
            num_parallel_workers (int): The number of concurrent workers. Default: None.
            is_training (bool): Specifies if is in training step. Default: True.

        Returns:
            data.ImageFolderDataset, the preprocessed ImageFolderDataset instance.

        Examples:
            >>> from tinyms.vision import ImageFolderTransform
            >>>
            >>> imagefolder_transform = ImageFolderTransform()
            >>> imagefolder_ds = imagefolder_transform.apply_ds(imagefolder_ds)
        '''
        if not isinstance(imagefolder_ds, ImageFolderDataset):
            raise TypeError("Input type should be ImageFolderDataset, got {}.".format(type(imagefolder_ds)))

        if is_training:
            trans_func = [self.random_crop_decode_resize, self.random_horizontal_flip]
        else:
            trans_func = [decode, self.resize, self.center_crop]
        trans_func += [self.normalize, hwc2chw]
        # apply transform functions on imagefolder dataset
        imagefolder_ds = super().apply_ds(imagefolder_ds, trans_func=trans_func, repeat_size=repeat_size,
                                          batch_size=batch_size, num_parallel_workers=num_parallel_workers)

        return imagefolder_ds


def _rand(a=0., b=1.):
    """Generate random."""
    return np.random.rand() * (b - a) + a


class VOCTransform(DatasetTransform):
    r'''
    VOC dataset transform class.

    Inputs:
        img (Union[numpy.ndarray, PIL.Image]): Image to be transformed in VOC-style.

    Outputs:
        numpy.ndarray, transformed image.

    Examples:
        >>> from PIL import Image
        >>> from tinyms.vision import VOCTransform
        >>>
        >>> voc_transform = VOCTransform()
        >>> img = Image.open('object_detection.jpg')
        >>> img = voc_transform(img)
    '''

    def __init__(self, configs=None):
        if configs:
            self.configs = configs
        else:
            self.configs = get_specified_config('VOCTransform')
        super().__init__(configs=self.configs)
        self.resize = Resize(self.configs['resize'])
        self.horizontal_flip = PILRandomHorizontalFlip(self.configs['horizontal_flip'])
        self.normalize = Normalize(self.configs['normalize']['mean'], self.configs['normalize']['std'])
        self.random_color_adjust = RandomColorAdjust(brightness=self.configs['random_color_adjust']['brightness'],
                                                     contrast=self.configs['random_color_adjust']['contrast'],
                                                     saturation=self.configs['random_color_adjust']['saturation'])

    def _preprocess_fn(self, image, boxes, labels, is_training=True):
        """Preprocess function for voc dataset."""
        def _random_sample_crop(image, boxes):
            """Random Crop the image and boxes"""
            height, width, _ = image.shape
            min_iou = np.random.choice([None, 0.1, 0.3, 0.5, 0.7, 0.9])
            if min_iou is None:
                return image, boxes
            # max trails (50)
            for _ in range(50):
                image_t = image
                w = _rand(0.3, 1.0) * width
                h = _rand(0.3, 1.0) * height
                # aspect ratio constraint b/t .5 & 2
                if h / w < 0.5 or h / w > 2:
                    continue
                left = _rand() * (width - w)
                top = _rand() * (height - h)
                rect = np.array([int(top), int(left), int(top + h), int(left + w)])
                overlap = jaccard_numpy(boxes, rect)
                # dropout some boxes
                drop_mask = overlap > 0
                if not drop_mask.any():
                    continue
                if overlap[drop_mask].min() < min_iou and overlap[drop_mask].max() > (min_iou + 0.2):
                    continue
                image_t = image_t[rect[0]:rect[2], rect[1]:rect[3], :]
                centers = (boxes[:, :2] + boxes[:, 2:4]) / 2.0
                m1 = (rect[0] < centers[:, 0]) * (rect[1] < centers[:, 1])
                m2 = (rect[2] > centers[:, 0]) * (rect[3] > centers[:, 1])
                # mask in that both m1 and m2 are true
                mask = m1 * m2 * drop_mask
                # have any valid boxes? try again if not
                if not mask.any():
                    continue
                # take only matching gt boxes
                boxes_t = boxes[mask, :].copy()
                boxes_t[:, :2] = np.maximum(boxes_t[:, :2], rect[:2])
                boxes_t[:, :2] -= rect[:2]
                boxes_t[:, 2:4] = np.minimum(boxes_t[:, 2:4], rect[2:4])
                boxes_t[:, 2:4] -= rect[:2]
                return image_t, boxes_t
            return image, boxes

        # Only perform resize operation of data evaluation step
        if not is_training:
            img_h, img_w, _ = image.shape
            image = self.resize(image)
            return image, np.array((img_h, img_w), dtype=np.float32), labels
        # Merge [x, y, w, h] and cls to [x, y, w, h, cls]
        boxes = np.hstack((boxes, labels)).astype(np.float32)
        # Change [x, y, w, h, cls] to [ymin, xmin, ymax, xmax, cls]
        boxes_yxyx = np.zeros_like(boxes)
        boxes_yxyx[:, 4] = boxes[:, 4]
        boxes_yxyx[:, [1, 0]] = boxes[:, [0, 1]]
        boxes_yxyx[:, [3, 2]] = boxes[:, [0, 1]] + boxes[:, [2, 3]]
        # Random crop image and bbox
        image, boxes_yxyx = _random_sample_crop(image, boxes_yxyx)
        # Resize image and bbox
        ih, iw, _ = image.shape
        image = self.resize(image)
        boxes_yxyx[:, [0, 2]] = boxes_yxyx[:, [0, 2]] / ih
        boxes_yxyx[:, [1, 3]] = boxes_yxyx[:, [1, 3]] / iw
        # Flip image and bbox or not
        flip = _rand() < .5
        if flip:
            image = np.asarray(self.horizontal_flip(Image.fromarray(image, mode='RGB')))
            boxes_yxyx[:, [1, 3]] = 1 - boxes_yxyx[:, [3, 1]]
        # When the channels of image is 1
        if len(image.shape) == 2:
            image = np.expand_dims(image, axis=-1)
            image = np.concatenate([image, image, image], axis=-1)

        boxes_yxyx, label, num_match = ssd_bboxes_encode(boxes_yxyx)
        return image, boxes_yxyx, label, num_match

    def __call__(self, img):
        if not isinstance(img, (np.ndarray, Image.Image)):
            raise TypeError("Input type should be numpy.ndarray or PIL.Image, got {}.".format(type(img)))
        img = self.resize(img)
        img = self.normalize(img)
        img = hwc2chw(img)

        return img

    def apply_ds(self, voc_ds, repeat_size=1, batch_size=32,
                 num_parallel_workers=None, is_training=True):
        r'''
        Apply preprocess operation on VOCDataset instance.

        Args:
            voc_ds (data.VOCDataset): VOCDataset instance.
            repeat_size (int): The repeat size of dataset. Default: 1.
            batch_size (int): Batch size. Default: 32.
            num_parallel_workers (int): The number of concurrent workers. Default: None.
            is_training (bool): Specifies if is in training step. Default: True.

        Returns:
            data.VOCDataset, the preprocessed VOCDataset instance.

        Examples:
            >>> from tinyms.vision import VOCTransform
            >>>
            >>> VOC_transform = VOCTransform()
            >>> voc_ds = voc_transform.apply_ds(voc_ds)
        '''
        if not isinstance(voc_ds, VOCDataset):
            raise TypeError("Input type should be VOCDataset, got {}.".format(type(voc_ds)))

        compose_map_func = (lambda image, boxes, labels: self._preprocess_fn(image, boxes, labels, is_training))
        if is_training:
            output_columns = ["image", "bbox", "label", "num_match"]
            trans_func = [self.random_color_adjust, self.normalize, hwc2chw]
        else:
            output_columns = ["image", "image_shape", "label"]
            trans_func = [self.normalize, hwc2chw]
        # apply transform functions on voc dataset
        voc_ds = voc_ds.map(operations=compose_map_func,
                            input_columns=["image", "bbox", "label"],
                            output_columns=output_columns,
                            column_order=output_columns,
                            num_parallel_workers=num_parallel_workers)
        voc_ds = super().apply_ds(voc_ds, trans_func=trans_func, repeat_size=repeat_size,
                                  batch_size=batch_size, num_parallel_workers=num_parallel_workers)

        return voc_ds

    def postprocess(self, input, image_shape, strategy='TOP1_CLASS'):
        r'''
        Apply postprocess operation for prediction result.

        Args:
            input (numpy.ndarray): Prediction result.
            image_shape (tuple): Image shape.
            strategy (str): Specifies the postprocess strategy. Default: TOP1_CLASS.

        Returns:
            dict, the postprocess result.
        '''
        if not isinstance(input, np.ndarray):
            raise TypeError("Input type should be numpy.ndarray, got {}.".format(type(input)))
        if not input.ndim == 3:
            raise TypeError("Input should be 3-D Numpy, got {}.".format(input.ndim))
        if not strategy == 'TOP1_CLASS':
            raise ValueError("Currently VOC transform only supports 'TOP1_CLASS' strategy!")

        pred_res = []
        pred_loc, pred_cls, pred_label = ssd_bboxes_filter(input[0, :, :4], input[0, :, 4:], image_shape)
        for loc, score, label in zip(pred_loc, pred_cls, pred_label):
            pred_res.append({
                'bbox': [loc[1], loc[0], loc[3] - loc[1], loc[2] - loc[0]],
                'score': score,
                'category_id': self.labels[label],
            })

        return pred_res


class ShanshuiTransform(VOCTransform):
    r'''
    Shanshui dataset transform class.

    Inputs:
        img (Union[numpy.ndarray, PIL.Image]): Image to be transformed in VOC-style.

    Outputs:
        numpy.ndarray, transformed image.

    Examples:
        >>> from PIL import Image
        >>> from tinyms.vision import ShanshuiTransform
        >>>
        >>> shanshui_transform = ShanshuiTransform()
        >>> img = Image.open('object_detection.jpg')
        >>> img = shanshui_transform(img)
    '''

    def __init__(self, configs=None):
        if configs:
            self.configs = configs
        else:
            self.configs = get_specified_config('ShanshuiTransform')
        super().__init__(configs=self.configs)

    def __call__(self, img):
        if not isinstance(img, (np.ndarray, Image.Image)):
            raise TypeError("Input type should be numpy.ndarray or PIL.Image, got {}.".format(type(img)))
        img = self.resize(img)
        img = self.normalize(img)
        img = hwc2chw(img)
        return img


class CycleGanDatasetTransform():
    r'''
    CycleGan dataset transform class.

    Inputs:
        img (Union[numpy.ndarray, PIL.Image]): Image to be transformed in city_scape.

    Outputs:
        numpy.ndarray, transformed image.

    Examples:
        >>> from PIL import Image
        >>> from tinyms.vision import CycleGanDatasetTransform
        >>>
        >>> cyclegan_transform = CycleGanDatasetTransform()
        >>> img = Image.open('object_detection.jpg')
        >>> img = cyclegan_transform(img)
    '''

    def __init__(self, configs=None):
        if configs:
            self.configs = configs
        else:
            self.configs = get_specified_config('CycleGanDatasetTransform')
        self.random_resized_crop = RandomResizedCrop(self.configs['random_resized_crop']['size'],
                                                     scale=tuple(self.configs['random_resized_crop']['scale']),
                                                     ratio=tuple(self.configs['random_resized_crop']['ratio']))
        self.random_horizontal_flip = RandomHorizontalFlip(prob=self.configs['random_horizontal_flip']['prob'])
        self.resize = Resize(self.configs['resize'])
        self.normalize = Normalize(mean=self.configs['normalize']['mean'], std=self.configs['normalize']['std'])

    def __call__(self, img):
        if not isinstance(img, (np.ndarray, Image.Image)):
            raise TypeError("Input type should be numpy.ndarray or PIL.Image, got {}.".format(type(img)))
        img = self.resize(img)
        img = self.normalize(img)
        img = hwc2chw(img)

        return img

    def apply_ds(self, gan_generator_ds, repeat_size=1, batch_size=1,
                 num_parallel_workers=1, shuffle=True, phase='train'):
        r'''
        Apply preprocess operation on GeneratorDataset instance.

        Args:
            gan_generator_ds (data.GeneratorDataset): GeneratorDataset instance.
            repeat_size (int): The repeat size of dataset. Default: 1.
            batch_size (int): Batch size. Default: 32.
            num_parallel_workers (int): The number of concurrent workers. Default: 1.
            shuffle (bool): Specifies if applying shuffle operation. Default: True.
            phase (str): Specifies the current phase. Default: train.

        Returns:
            data.GeneratorDataset, the preprocessed GeneratorDataset instance.

        Examples:
            >>> from tinyms.vision import CycleGanDatasetTransform
            >>>
            >>> cyclegan_transform = CycleGanDatasetTransform()
            >>> gan_generator_ds = cyclegan_transform.apply_ds(gan_generator_ds)

        Raises:
            TypeError: If `gan_generator_ds` is not instance of GeneratorDataset.
        '''
        if not isinstance(gan_generator_ds, GeneratorDataset):
            raise TypeError("Input type should be GeneratorDataset, got {}.".format(type(gan_generator_ds)))

        trans_func = []
        if phase == 'train':
            if shuffle:
                trans_func += [self.random_resized_crop, self.random_horizontal_flip, self.normalize, hwc2chw]
            else:
                trans_func += [self.resize, self.normalize, hwc2chw]

            # apply transform functions on gan_generator_ds dataset
            gan_generator_ds = gan_generator_ds.map(operations=trans_func,
                                                    input_columns=["image_A"],
                                                    num_parallel_workers=num_parallel_workers)
            gan_generator_ds = gan_generator_ds.map(operations=trans_func,
                                                    input_columns=["image_B"],
                                                    num_parallel_workers=num_parallel_workers)
        else:
            trans_func += [self.resize, self.normalize, hwc2chw]
            gan_generator_ds = gan_generator_ds.map(operations=trans_func,
                                                    input_columns=["image"],
                                                    num_parallel_workers=num_parallel_workers)
        gan_generator_ds = gan_generator_ds.batch(batch_size, drop_remainder=True)
        gan_generator_ds = gan_generator_ds.repeat(repeat_size)
        return gan_generator_ds


mnist_transform = MnistTransform()
cifar10_transform = Cifar10Transform()
imagefolder_transform = ImageFolderTransform()
voc_transform = VOCTransform()
shanshui_tranform = ShanshuiTransform()
cyclegan_transform = CycleGanDatasetTransform()


# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Primitives module. Operators can be used in the construct function of Layer.

Examples:
    >>> import tinyms as ts
    >>> from tinyms.primitives import tensor_add
    >>>
    >>> x = ts.ones([2, 3])
    >>> y = ts.ones([2, 3])
    >>> print(tensor_add(x, y))
    [[2. 2. 2.]
    [2. 2. 2.]]
"""
from mindspore.ops import composite, operations, functional
from mindspore.ops.composite import *
from mindspore.ops.operations import *
from mindspore.ops.functional import *

__all__ = []
__all__.extend(composite.__all__)
__all__.extend(operations.__all__)
__all__.extend(functional.__all__)



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Download or extract file."""

import os
import shutil
import hashlib
import errno
import stat
import urllib
from urllib.request import urlretrieve, HTTPError, URLError
import ssl


ssl._create_default_https_context = ssl._create_unverified_context
REAL_PATH = os.path.split(os.path.realpath(__file__))[0]
MAX_FILE_SIZE = 5 * 1024 * 1024 * 1024  # 5GB
SUFFIX_LIST = ['.ckpt', 'mindir', '.onnx']


def handle_remove_read_only(func, path, exc):
    exc_value = exc[1]
    if func in (os.rmdir, os.remove, os.unlink) and exc_value.errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # 0777
        func(path)


def url_exist(url):
    """
    Whether the url exist.
    """
    opener = urllib.request.build_opener()
    opener.addheaders = [('User-Agent', 'Mozilla/5.0')]
    try:
        opener.open(url)
        return True
    except HTTPError as e:
        print(e.code)
    except URLError as e:
        print(e.reason)
    return False


def _remove_path_if_exists(path):
    if os.path.exists(path):
        if os.path.isfile(path):
            os.remove(path)
        else:
            shutil.rmtree(path, ignore_errors=False, onerror=handle_remove_read_only)


def _create_path_if_not_exists(path):
    if not os.path.exists(path):
        if os.path.isfile(path):
            os.remove(path)
        else:
            os.makedirs(path)


def download_file_from_url(url, hash_sha256=None, save_path='.'):
    """
    download checkpoint weight from giving url.

    Args:
       url(string): checkpoint url path.
       hash_sha256(string): checkpoint file sha256.
       save_path(string): checkpoint download save path.

    Returns:
       string.
    """

    def reporthook(a, b, c):
        percent = a * b * 100.0 / c
        percent = 100 if percent > 100 else percent
        if c > 0:
            print("\rDownloading...%5.1f%%" % percent, end="")

    def sha256sum(file_name, hash_sha256):
        fp = open(file_name, 'rb')
        content = fp.read()
        fp.close()
        m = hashlib.sha256()
        m.update(content)
        download_sha256 = m.hexdigest()
        return download_sha256 == hash_sha256

    _create_path_if_not_exists(os.path.realpath(save_path))
    ckpt_name = os.path.basename(url.split("/")[-1])
    # identify file exist or not
    file_path = os.path.join(save_path, ckpt_name)
    if os.path.isfile(file_path):
        if hash_sha256 and sha256sum(file_path, hash_sha256):
            print('File already exists!')
            return file_path
        print('File already exists, but sha256 checking failed. Will download again')

    _remove_path_if_exists(file_path)

    # download the checkpoint file
    print('Downloading data from url {}'.format(url))
    try:
        opener = urllib.request.build_opener()
        opener.addheaders = [('User-Agent', 'Mozilla/5.0')]
        urllib.request.install_opener(opener)
        urlretrieve(url, file_path, reporthook=reporthook)
    except HTTPError as e:
        raise Exception(e.code, e.msg, url)
    except URLError as e:
        raise Exception(e.errno, e.reason, url)
    print('\nDownload finished!')

    # Check file integrity
    if hash_sha256:
        result = sha256sum(file_path, hash_sha256)
        if not result:
            raise Exception('INTEGRITY ERROR: File: {} is not integral'.format(file_path))

    # Check if file size over MAX_FILE_SIZE
    file_size = os.path.getsize(file_path)
    if file_size > MAX_FILE_SIZE:
        os.remove(file_path)
        raise Exception('SIZE ERROR: Download file is too large,'
                        'the max size is {}Mb'.format(MAX_FILE_SIZE / 1024 / 1024))

    # Check file type
    suffix = os.path.splitext(file_path)[1]
    if suffix not in SUFFIX_LIST:
        os.remove(file_path)
        raise Exception('SUFFIX ERROR: File: {} with suffix: {} '
                        'can not be recognized'.format(file_path, suffix))
    return file_path



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Introduction to data/loader:

data/loader supports various formats of datasets, including ImageNet, TFData,
MNIST, Cifar10/100, Manifest, MindRecord, etc. This module could load data in
high performance and parse data precisely. It also provides the following
operations for users to preprocess data: shuffle, batch, repeat, map, and zip.
"""
import os
import random
import numpy as np
import math
import gensim
import codecs
import collections
import pickle

from itertools import chain
from tqdm import tqdm

from mindspore.dataset import engine
from mindspore.dataset.engine import *
from mindspore.mindrecord import FileWriter
from .utils import generate_image_list, load_img


common_dataset = ['UnalignedDataset', 'GanImageFolderDataset', 'ImdbDataset', 'BertDataset',
                  'KaggleDisplayAdvertisingDataset']
common_sampler = ['DistributedSampler']

__all__ = common_dataset + common_sampler
__all__.extend(engine.__all__)

random.seed(1)


class UnalignedDataset:
    """
    This dataset class can load unaligned/unpaired datasets.

    Args:
        dataset_path (str): The path of images (should have subfolders trainA, trainB, testA, testB, etc).
        phase (str): Train or test. It requires two directories in dataset_path, like trainA and trainB to.
            host training images from domain A '{dataset_path}/trainA' and from domain B '{dataset_path}/trainB'
            respectively.
        max_dataset_size (int): Maximum number of return image paths.

    Returns:
        Two domain image path list.
    """

    def __init__(self, dataset_path, phase, max_dataset_size=float("inf"), shuffle=True):
        self.dir_A = os.path.join(dataset_path, phase + 'A')
        self.dir_B = os.path.join(dataset_path, phase + 'B')

        self.A_paths = sorted(generate_image_list(self.dir_A,
                                                  max_dataset_size))  # load images from '/path/to/data/trainA'
        self.B_paths = sorted(generate_image_list(self.dir_B,
                                                  max_dataset_size))    # load images from '/path/to/data/trainB'
        self.A_size = len(self.A_paths)  # get the size of dataset A
        self.B_size = len(self.B_paths)  # get the size of dataset B
        self.shuffle = shuffle

    def __getitem__(self, index):
        index_B = index % self.B_size
        if index % max(self.A_size, self.B_size) == 0 and self.shuffle:
            random.shuffle(self.A_paths)
            index_B = random.randint(0, self.B_size - 1)
        A_path = self.A_paths[index % self.A_size]
        B_path = self.B_paths[index_B]
        A_img = np.array(load_img(A_path))
        B_img = np.array(load_img(B_path))

        return A_img, B_img

    def __len__(self):
        return max(self.A_size, self.B_size)


class GanImageFolderDataset:
    """
    This dataset class can load images from image folder.

    Args:
        dataset_path (str): '{dataset_path}/testA', '{dataset_path}/testB', etc.
        max_dataset_size (int): Maximum number of return image paths.

    Returns:
        Image path list.
    """

    def __init__(self, dataset_path, max_dataset_size=float("inf")):
        self.dataset_path = dataset_path
        self.paths = sorted(generate_image_list(dataset_path, max_dataset_size))
        self.size = len(self.paths)

    def __getitem__(self, index):
        img_path = self.paths[index % self.size]
        img = np.array(load_img(img_path))

        return img, os.path.split(img_path)[1]

    def __len__(self):
        return self.size


class BertDataset:
    """
    This dataset class can load bert from data folder.

    Args:
        data_dir (str): '{data_dir}/result1.tfrecord', '{data_dir}/result2.tfrecord', etc.
        num_parallel_workers (int): The number of concurrent workers. Default: None.
        shuffle (Union[bool, Shuffle level], optional): Perform reshuffling of the data every epoch
            (default=Shuffle.GLOBAL).
            If shuffle is False, no shuffling will be performed;
            If shuffle is True, the behavior is the same as setting shuffle to be Shuffle.GLOBAL
            Otherwise, there are two levels of shuffling:

            - Shuffle.GLOBAL: Shuffle both the files and samples.

            - Shuffle.FILES: Shuffle files only.
        schema (Union[str, Schema], optional): Path to the JSON schema file or schema object (default=None).
            If the schema is not provided, the meta data from the TFData file is considered the schema.

    Examples:
        >>> from tinyms.data import BertDataset
        >>>
        >>> bert_ds = BertDataset('data')
    """

    def __init__(self, data_dir, schema_dir=None, shuffle=True, num_parallel_workers=None):
        files = os.listdir(data_dir)

        self.data_files = []
        for file_name in files:
            if "tfrecord" in file_name:
                self.data_files.append(os.path.join(data_dir, file_name))
        self.data_set = TFRecordDataset(self.data_files, schema_dir if schema_dir != "" else None,
                                      columns_list=["input_ids", "input_mask", "segment_ids", "next_sentence_labels",
                                                    "masked_lm_positions", "masked_lm_ids", "masked_lm_weights"],
                                      shuffle=shuffle,
                                      shard_equal_rows=True,
                                      num_parallel_workers=num_parallel_workers)


class ImdbDataset:
    """
    parse aclImdb data to features and labels.
    sentence->tokenized->encoded->padding->features

    Args:
        imdb_path (str): The path where the aclImdb dataset stored.
        glove_path (str): The path where the GloVe stored.
        embed_size (int): Embed_size. Default: 300.

    Examples:
        >>> from tinyms.data import ImdbDataset
        >>>
        >>> imdb_ds = ImdbDataset('./aclImdb', './glove')
    """

    def __init__(self, imdb_path, glove_path, embed_size=300):
        self.__segs = ['train', 'test']
        self.__label_dic = {'pos': 1, 'neg': 0}
        self.__imdb_path = imdb_path
        self.__glove_dim = embed_size
        self.__glove_file = os.path.join(glove_path, 'glove.6B.' + str(self.__glove_dim) + 'd.txt')

        # properties
        self.__imdb_datas = {}
        self.__features = {}
        self.__labels = {}
        self.__vacab = {}
        self.__word2idx = {}
        self.__weight_np = {}
        self.__wvmodel = None
        self.parse()

    def parse(self):
        """
        parse imdb data to memory
        """
        self.__wvmodel = gensim.models.KeyedVectors.load_word2vec_format(self.__glove_file, binary=False)

        for seg in self.__segs:
            self.__parse_imdb_datas(seg)
            self.__parse_features_and_labels(seg)
            self.__gen_weight_np(seg)

    def __parse_imdb_datas(self, seg):
        """
        load data from txt
        """
        data_lists = []
        for label_name, label_id in self.__label_dic.items():
            sentence_dir = os.path.join(self.__imdb_path, seg, label_name)
            for file in os.listdir(sentence_dir):
                with open(os.path.join(sentence_dir, file), mode='r', encoding='utf8') as f:
                    sentence = f.read().replace('\n', '')
                    data_lists.append([sentence, label_id])
        self.__imdb_datas[seg] = data_lists

    def __parse_features_and_labels(self, seg):
        """
        parse features and labels
        """
        features = []
        labels = []
        for sentence, label in self.__imdb_datas[seg]:
            features.append(sentence)
            labels.append(label)

        self.__features[seg] = features
        self.__labels[seg] = labels

        # update feature to tokenized
        self.__updata_features_to_tokenized(seg)
        # parse vacab
        self.__parse_vacab(seg)
        # encode feature
        self.__encode_features(seg)
        # padding feature
        self.__padding_features(seg)

    def __updata_features_to_tokenized(self, seg):
        tokenized_features = []
        for sentence in self.__features[seg]:
            tokenized_sentence = [word.lower() for word in sentence.split(" ")]
            tokenized_features.append(tokenized_sentence)
        self.__features[seg] = tokenized_features

    def __parse_vacab(self, seg):
        # vocab
        tokenized_features = self.__features[seg]
        vocab = set(chain(*tokenized_features))
        self.__vacab[seg] = vocab

        # word_to_idx looks like {'hello': 1, 'world':111, ... '<unk>': 0}
        word_to_idx = {word: i + 1 for i, word in enumerate(vocab)}
        word_to_idx['<unk>'] = 0
        self.__word2idx[seg] = word_to_idx

    def __encode_features(self, seg):
        """ encode word to index """
        word_to_idx = self.__word2idx['train']
        encoded_features = []
        for tokenized_sentence in self.__features[seg]:
            encoded_sentence = []
            for word in tokenized_sentence:
                encoded_sentence.append(word_to_idx.get(word, 0))
            encoded_features.append(encoded_sentence)
        self.__features[seg] = encoded_features

    def __padding_features(self, seg, maxlen=500, pad=0):
        """ pad all features to the same length """
        padded_features = []
        for feature in self.__features[seg]:
            if len(feature) >= maxlen:
                padded_feature = feature[:maxlen]
            else:
                padded_feature = feature
                while len(padded_feature) < maxlen:
                    padded_feature.append(pad)
            padded_features.append(padded_feature)
        self.__features[seg] = padded_features

    def __gen_weight_np(self, seg):
        """
        generate weight by gensim
        """
        weight_np = np.zeros((len(self.__word2idx[seg]), self.__glove_dim), dtype=np.float32)
        for word, idx in self.__word2idx[seg].items():
            if word not in self.__wvmodel:
                continue
            word_vector = self.__wvmodel.get_vector(word)
            weight_np[idx, :] = word_vector

        self.__weight_np[seg] = weight_np

    def get_datas(self, seg):
        """
        get features, labels, and weight by gensim.
        """
        features = np.array(self.__features[seg]).astype(np.int32)
        labels = np.array(self.__labels[seg]).astype(np.int32)
        weight = np.array(self.__weight_np[seg])
        return features, labels, weight

    def convert_to_mindrecord(self, preprocess_path, shard_num=1):
        """
        convert imdb dataset to mindrecoed dataset
        """

        if not os.path.exists(preprocess_path):
            print(f"preprocess path {preprocess_path} is not exist")
            os.makedirs(preprocess_path)

        train_features, train_labels, train_weight_np = self.get_datas('train')
        self._convert_to_mindrecord(preprocess_path, train_features, train_labels, train_weight_np, shard_num=shard_num)

        test_features, test_labels, _ = self.get_datas('test')
        self._convert_to_mindrecord(preprocess_path, test_features, test_labels, training=False, shard_num=shard_num)

    def _convert_to_mindrecord(self, data_home, features, labels, weight_np=None, training=True, shard_num=1):
        """
        convert imdb dataset to mindrecoed dataset
        """
        if weight_np is not None:
            np.savetxt(os.path.join(data_home, 'weight.txt'), weight_np)

        # write mindrecord
        schema_json = {"id": {"type": "int32"},
                       "label": {"type": "int32"},
                       "feature": {"type": "int32", "shape": [-1]}}

        data_dir = os.path.join(data_home, "aclImdb_train.mindrecord")
        if not training:
            data_dir = os.path.join(data_home, "aclImdb_test.mindrecord")

        def get_imdb_data(features, labels):
            data_list = []
            for i, (label, feature) in enumerate(zip(labels, features)):
                data_json = {"id": i,
                             "label": int(label),
                             "feature": feature.reshape(-1)}
                data_list.append(data_json)
            return data_list

        writer = FileWriter(data_dir, shard_num=shard_num)
        data = get_imdb_data(features, labels)
        writer.add_schema(schema_json, "nlp_schema")
        writer.add_index(["id", "label"])
        writer.write_raw_data(data)
        writer.commit()


class KaggleDisplayAdvertisingDataset:
    """
    parse aclImdb data to features and labels.
    sentence->tokenized->encoded->padding->features

    Args:
        data_dir (str): The path where the uncompressed dataset stored.
        num_parallel_workers (int): The number of concurrent workers. Default: None.
        shuffle (bol): Whether the dataset needs to be shuffled. Default: True.

    Examples:
        >>> from tinyms.data import KaggleDisplayAdvertisingDataset
        >>>
        >>> kaggle_display_advertising_ds = KaggleDisplayAdvertisingDataset('data')
        >>> kaggle_display_advertising_ds.stats_data()
        >>> kaggle_display_advertising_ds.convert_to_mindrecord()
        >>> train_ds = kaggle_display_advertising_ds.load_mindreocrd_dataset(usage='train')
        >>> test_ds = kaggle_display_advertising_ds.load_mindreocrd_dataset(usage='test')
    """
    def __init__(self, data_dir, num_parallel_workers=None, shuffle=True):
        self.data_dir = data_dir
        self.dense_dim = 13
        self.slot_dim = 26
        self.field_size = 39  # dense_dim + slot_dim
        self.skip_id_convert = False
        self.train_line_count = 45840617
        self.test_size = 0.1
        self.seed = 20191005
        self.line_per_sample = 1000
        self.epochs = 1
        self.num_parallel_workers = num_parallel_workers
        self._check_num_parallel_workers()
        self.shuffle = shuffle

        self.__init_stats()
        self.__init_mindrecord()

    def __check_mindrecord_dir(self):
        if os.path.exists(self.mindrecord_dir):
            print("mindrecord directory: {} exists! we will use it to save or read mindrecord dataset.".
                  format(self.mindrecord_dir), flush=True)
        else:
            print("create mindrecord directory: {} to save or read mindrecord dataset.".
                  format(self.mindrecord_dir), flush=True)
            os.makedirs(self.mindrecord_dir)

    def _check_num_parallel_workers(self):
        # use multiprocessing to get cpu count
        import multiprocessing
        cpu_count = multiprocessing.cpu_count()

        if self.num_parallel_workers is not None:
            if cpu_count < self.num_parallel_workers:
                raise ValueError("num_parallel_workers: {} is bigger than cpu count: {}!".format(
                    self.num_parallel_workers, cpu_count))

    def __check_stats_dict_dir(self):
        if os.path.exists(self.stats_dict_dir):
            print("stats dict directory: {} exists! we will use it to save or read stats dict.".
                  format(self.stats_dict_dir), flush=True)
        else:
            print("create stats dict directory: {} to save or read stats dict.".
                  format(self.stats_dict_dir), flush=True)
            os.makedirs(self.stats_dict_dir)

    def __init_mindrecord(self):
        """
        init mindrecord
        """
        self.mindrecord_dir = os.path.join(self.data_dir, "mindrecord")
        self.__check_mindrecord_dir()

    def __init_stats(self):
        """
        init stats values
        """
        self.val_cols = ["val_{}".format(i + 1) for i in range(self.dense_dim)]
        self.cat_cols = ["cat_{}".format(i + 1) for i in range(self.slot_dim)]
        self.val_min_dict = {col: 0 for col in self.val_cols}
        self.val_max_dict = {col: 0 for col in self.val_cols}
        self.cat_count_dict = {col: collections.defaultdict(int) for col in self.cat_cols}
        self.oov_prefix = "OOV"
        self.cat2id_dict = {}
        self.cat2id_dict.update({col: i for i, col in enumerate(self.val_cols)})
        self.cat2id_dict.update(
            {self.oov_prefix + col: i + len(self.val_cols) for i, col in enumerate(self.cat_cols)})
        self.stats_dict_dir = os.path.join(self.data_dir, "stats_dict/")
        self.__check_stats_dict_dir()

    def __stats_vals(self, val_list):
        """
        handling weights column
        """
        assert len(val_list) == len(self.val_cols)
        for i, val in enumerate(val_list):
            key = self.val_cols[i]
            if val != "":
                if float(val) > self.val_max_dict[key]:
                    self.val_max_dict[key] = float(val)
                if float(val) < self.val_min_dict[key]:
                    self.val_min_dict[key] = float(val)

    def __stats_cats(self, cat_list):
        """
        handling cats column
        """
        assert len(cat_list) == len(self.cat_cols)
        for i, cat in enumerate(cat_list):
            key = self.cat_cols[i]
            self.cat_count_dict[key][cat] += 1

    def __save_stats_dict(self):
        """
        save stats dict
        """
        with open(os.path.join(self.stats_dict_dir, "val_max_dict.pkl"), "wb") as f:
            pickle.dump(self.val_max_dict, f)
        with open(os.path.join(self.stats_dict_dir, "val_min_dict.pkl"), "wb") as f:
            pickle.dump(self.val_min_dict, f)
        with open(os.path.join(self.stats_dict_dir, "cat_count_dict.pkl"), "wb") as f:
            pickle.dump(self.cat_count_dict, f)

    def __load_stats_dict(self,):
        """
        load stats dict
        """
        with open(os.path.join(self.stats_dict_dir, "val_max_dict.pkl"), "rb") as f:
            self.val_max_dict = pickle.load(f)
        with open(os.path.join(self.stats_dict_dir, "val_min_dict.pkl"), "rb") as f:
            self.val_min_dict = pickle.load(f)
        with open(os.path.join(self.stats_dict_dir, "cat_count_dict.pkl"), "rb") as f:
            self.cat_count_dict = pickle.load(f)
        print("val_max_dict.items()[:50]:{}".format(list(self.val_max_dict.items())), flush=True)
        print("val_min_dict.items()[:50]:{}".format(list(self.val_min_dict.items())), flush=True)

    def __get_cat2id(self, threshold=100):
        for key, cat_count_d in self.cat_count_dict.items():
            new_cat_count_d = dict(filter(lambda x: x[1] > threshold, cat_count_d.items()))
            for cat_str, _ in new_cat_count_d.items():
                self.cat2id_dict[key + "_" + cat_str] = len(self.cat2id_dict)
        print("cat2id_dict.size:{}".format(len(self.cat2id_dict)), flush=True)
        print("cat2id.dict.items()[:50]:{}".format(list(self.cat2id_dict.items())[:50]), flush=True)

    def __map_cat2id(self, values, cats):
        """
        category value to id value
        """
        id_list = []
        weight_list = []
        for i, val in enumerate(values):
            if val == "":
                id_list.append(i)
                weight_list.append(0)
            else:
                key = "val_{}".format(i + 1)
                id_list.append(self.cat2id_dict[key])
                max_v = float(self.val_max_dict["val_{}".format(i + 1)])
                minmax_scale_value = float(val) * 1.0 / max_v
                weight_list.append(minmax_scale_value)

        for i, cat_str in enumerate(cats):
            key = "cat_{}".format(i + 1) + "_" + cat_str
            if key in self.cat2id_dict:
                if self.skip_id_convert is True:
                    # For the synthetic data, if the generated id is between [0, max_vcoab], but the num examples is l
                    # ess than vocab_size/ slot_nums the id will still be converted to [0, real_vocab], where real_vocab
                    # the actually the vocab size, rather than the max_vocab. So a simple way to alleviate this
                    # problem is skip the id convert, regarding the synthetic data id as the final id.
                    id_list.append(cat_str)
                else:
                    id_list.append(self.cat2id_dict[key])
            else:
                id_list.append(self.cat2id_dict[self.oov_prefix + "cat_{}".format(i + 1)])
            weight_list.append(1.0)
        return id_list, weight_list

    def stats_data(self):
        """
        stats data
        """
        num_splits = self.dense_dim + self.slot_dim + 1
        error_stat_lines_num = []

        train_file_path = os.path.join(self.data_dir, "train.txt")
        with codecs.open(train_file_path, encoding="utf8", buffering=32*1024*1024) as f:
            t_f = tqdm(f, total=self.train_line_count)
            t_f.set_description("Processing StatsData")
            num_line = 0
            for line in t_f:
                num_line += 1
                line = line.strip("\n")
                items = line.split("\t")
                if len(items) != num_splits:
                    error_stat_lines_num.append(num_line)
                    # print("Found line length: {}, suppose to be {}, the line is {}".format(
                    #     len(items), num_splits, line))
                    continue
                # if num_line % 1000000 == 0:
                #     print("Have handled {}w lines.".format(num_line // 10000))
                values = items[1: self.dense_dim + 1]
                cats = items[self.dense_dim + 1:]
                assert len(values) == self.dense_dim, "values.size: {}".format(len(values))
                assert len(cats) == self.slot_dim, "cats.size: {}".format(len(cats))
                self.__stats_vals(values)
                self.__stats_cats(cats)
        self.__save_stats_dict()

        error_stat_path = os.path.join(self.data_dir, "error_stat_lines_num.npy")
        np.save(error_stat_path, error_stat_lines_num)

    def convert_to_mindrecord(self):
        test_size = int(self.train_line_count * self.test_size)
        all_indices = [i for i in range(self.train_line_count)]
        np.random.seed(self.seed)
        np.random.shuffle(all_indices)
        test_indices_set = set(all_indices[:test_size])

        train_data_list = []
        test_data_list = []
        ids_list = []
        wts_list = []
        label_list = []

        schema = {
            "label": {"type": "float32", "shape": [-1]},
            "feat_vals": {"type": "float32", "shape": [-1]},
            "feat_ids": {"type": "int32", "shape": [-1]}
        }

        train_writer = FileWriter(os.path.join(self.mindrecord_dir, "train_input_part.mindrecord"), 21)
        test_writer = FileWriter(os.path.join(self.mindrecord_dir, "test_input_part.mindrecord"), 3)
        train_writer.add_schema(schema, "CRITEO_TRAIN")
        test_writer.add_schema(schema, "CRITEO_TEST")

        part_rows = 2000000
        num_splits = self.dense_dim + self.slot_dim + 1
        error_conv_lines_num = []

        train_file_path = os.path.join(self.data_dir, "train.txt")
        with codecs.open(train_file_path, encoding="utf8", buffering=32*1024*1024) as f:
            t_f = tqdm(f, total=self.train_line_count)
            t_f.set_description("Processing Convert2MR")
            num_line = 0
            train_part_number = 0
            test_part_number = 0
            for line in t_f:
                num_line += 1
                # if num_line % 1000000 == 0:
                #     print("Converting to MindRecord. Have handle {}w lines.".format(num_line // 10000), flush=True)
                line = line.strip("\n")
                items = line.split("\t")
                if len(items) != num_splits:
                    error_conv_lines_num.append(num_line)
                    continue
                label = float(items[0])
                values = items[1: 1 + self.dense_dim]
                cats = items[1+self.dense_dim:]

                assert len(values) == self.dense_dim, "values.size: {}".format(len(values))
                assert len(cats) == self.slot_dim, "cats.size: {}".format(len(cats))

                ids, wts = self.__map_cat2id(values, cats)
                ids_list.extend(ids)
                wts_list.extend(wts)
                label_list.append(label)

                if num_line % self.line_per_sample == 0:
                    if num_line not in test_indices_set:
                        train_data_list.append({"feat_ids": np.array(ids_list, dtype=np.int32),
                                                "feat_vals": np.array(wts_list, dtype=np.float32),
                                                "label": np.array(label_list, dtype=np.float32)
                                                })
                    else:
                        test_data_list.append({"feat_ids": np.array(ids_list, dtype=np.int32),
                                               "feat_vals": np.array(wts_list, dtype=np.float32),
                                               "label": np.array(label_list, dtype=np.float32)
                                               })
                    if train_data_list and len(train_data_list) % part_rows == 0:
                        train_writer.write_raw_data(train_data_list)
                        train_data_list.clear()
                        train_part_number += 1

                    if test_data_list and len(test_data_list) % part_rows == 0:
                        test_writer.write_raw_data(test_data_list)
                        test_data_list.clear()
                        test_part_number += 1

                    ids_list.clear()
                    wts_list.clear()
                    label_list.clear()
            if train_data_list:
                train_writer.write_raw_data(train_data_list)
            if test_data_list:
                test_writer.write_raw_data(test_data_list)
        train_writer.commit()
        test_writer.commit()

        error_stat_path = os.path.join(self.data_dir, "error_conv_lines_num.npy")
        np.save(error_stat_path, error_conv_lines_num)

    def load_mindreocrd_dataset(self, usage='train', batch_size=1000):
        """
        load mindrecord dataset.
        Args:
            usage (str): Dataset mode. Default: 'train'.
            batch_size (int): batch size. Default: 1000.

        Returns:
            MindDataset
        """
        if usage == 'train':
            train_mode = True
        else:
            train_mode = False

        prefix_file_name = 'train_input_part.mindrecord' if train_mode else 'test_input_part.mindrecord'
        suffix_file_name = '00' if train_mode else '0'
        dataset_path = os.path.join(self.mindrecord_dir, prefix_file_name + suffix_file_name)

        dataset = MindDataset(dataset_path,
                              columns_list=['feat_ids', 'feat_vals', 'label'],
                              num_parallel_workers=self.num_parallel_workers,
                              shuffle=self.shuffle, num_shards=None, shard_id=None)

        dataset = dataset.batch(int(batch_size / self.line_per_sample), drop_remainder=True)
        dataset = dataset.map(operations=(lambda x, y, z: (np.array(x).flatten().reshape(batch_size, 39),
                                                           np.array(y).flatten().reshape(batch_size, 39),
                                                           np.array(z).flatten().reshape(batch_size, 1))),
                              input_columns=['feat_ids', 'feat_vals', 'label'],
                              column_order=['feat_ids', 'feat_vals', 'label'],
                              num_parallel_workers=self.num_parallel_workers)
        dataset = dataset.repeat(self.epochs)
        return dataset


class DistributedSampler:
    """
    Distributed sampler.

    Args:
        dataset_size (int): Dataset list length
        num_replicas (int): Replicas num.
        rank (int): Device rank.
        shuffle (bool): Whether the dataset needs to be shuffled. Default: True.

    Returns:
        DistributedSampler instance.
    """

    def __init__(self, dataset_size, num_replicas=None, rank=None, shuffle=True):
        if num_replicas is None:
            print("***********Setting world_size to 1 since it is not passed in ******************")
            num_replicas = 1
        if rank is None:
            print("***********Setting rank to 0 since it is not passed in ******************")
            rank = 0
        self.dataset_size = dataset_size
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        self.num_samples = int(math.ceil(dataset_size * 1.0 / self.num_replicas))
        self.total_size = self.num_samples * self.num_replicas
        self.shuffle = shuffle

    def __iter__(self):
        # deterministically shuffle based on epoch
        if self.shuffle:
            indices = np.random.RandomState(seed=self.epoch).permutation(self.dataset_size)
            # np.array type. number from 0 to len(dataset_size)-1, used as index of dataset
            indices = indices.tolist()
            self.epoch += 1
            # change to list type
        else:
            indices = list(range(self.dataset_size))

        # add extra samples to make it evenly divisible
        indices += indices[:(self.total_size - len(indices))]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank:self.total_size:self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Cycle GAN Common network."""
from tinyms.initializers import initializer, Normal, XavierUniform
from tinyms import layers


def init_weights(net, init_type='normal', init_gain=0.02):
    """
    Initialize network weights.

    Args:
        net (layers.Layer): Network to be initialized
        init_type (str): The name of an initialization method: normal | xavier.
        init_gain (float): Gain factor for normal and xavier.

    """
    for _, layer in net.cells_and_names():
        if isinstance(layer, (layers.Conv2d, layers.Conv2dTranspose)):
            if init_type == 'normal':
                layer.weight.set_data(initializer(Normal(init_gain), layer.weight.shape))
            elif init_type == 'xavier':
                layer.weight.set_data(initializer(XavierUniform(init_gain), layer.weight.shape))
            elif init_type == 'constant':
                layer.weight.set_data(initializer(0.001, layer.weight.shape))
            else:
                raise NotImplementedError('initialization method [%s] is not implemented' % init_type)
        elif isinstance(layer, layers.BatchNorm2d):
            layer.gamma.set_data(initializer('ones', layer.gamma.shape))
            layer.beta.set_data(initializer('zeros', layer.beta.shape))


class ConvNormReLU(layers.Layer):
    """
    Convolution fused with BatchNorm/InstanceNorm and ReLU/LackyReLU block definition.

    Args:
        in_planes (int): Input channel.
        out_planes (int): Output channel.
        kernel_size (int): Input kernel size. Default: 4.
        stride (int): Stride size for the first convolutional layer. Default: 2.
        alpha (float): Slope of LackyReLU. Default: 0.2.
        norm_mode (str): Specifies norm method. The optional values are "batch", "instance".
        pad_mode (str): Specifies padding mode. The optional values are "CONSTANT", "REFLECT", "SYMMETRIC".
            Default: "CONSTANT".
        use_relu (bool): Use relu or not. Default: True.
        padding (int): Pad size, if it is None, it will calculate by kernel_size. Default: None.

    Returns:
        Tensor, output tensor.
    """
    def __init__(self,
                 in_planes,
                 out_planes,
                 kernel_size=4,
                 stride=2,
                 alpha=0.2,
                 norm_mode='batch',
                 pad_mode='CONSTANT',
                 use_relu=True,
                 padding=None):
        super(ConvNormReLU, self).__init__()
        self.norm = layers.BatchNorm2d(out_planes)
        if norm_mode == 'instance':
            # Use BatchNorm2d with batchsize=1, affine=False, training=True instead of InstanceNorm2d
            norm = layers.BatchNorm2d(out_planes, affine=False)
        has_bias = (norm_mode == 'instance')
        if padding is None:
            padding = (kernel_size - 1) // 2
        if pad_mode == 'CONSTANT':
            conv = layers.Conv2d(in_planes, out_planes, kernel_size, stride, pad_mode='pad',
                                 has_bias=has_bias, padding=padding)
            layer_list = [conv, norm]
        else:
            paddings = ((0, 0), (0, 0), (padding, padding), (padding, padding))
            pad = layers.Pad(paddings=paddings, mode=pad_mode)
            conv = layers.Conv2d(in_planes, out_planes, kernel_size, stride, pad_mode='pad', has_bias=has_bias)
            layer_list = [pad, conv, norm]
        if use_relu:
            relu = layers.ReLU()
            if alpha > 0:
                relu = layers.LeakyReLU(alpha)
            layer_list.append(relu)
        self.features = layers.SequentialLayer(layer_list)

    def construct(self, x):
        output = self.features(x)
        return output


class ConvTransposeNormReLU(layers.Layer):
    """
    ConvTranspose2d fused with BatchNorm/InstanceNorm and ReLU/LackyReLU block definition.

    Args:
        in_planes (int): Input channel.
        out_planes (int): Output channel.
        kernel_size (int): Input kernel size. Default: 4.
        stride (int): Stride size for the first convolutional layer. Default: 2.
        alpha (float): Slope of LackyReLU. Default: 0.2.
        norm_mode (str): Specifies norm method. The optional values are "batch", "instance".
        pad_mode (str): Specifies padding mode. The optional values are "CONSTANT", "REFLECT", "SYMMETRIC".
                        Default: "CONSTANT".
        use_relu (bool): Use relu or not. Default: True.
        padding (int): The pad size, if it is None, it will calculate by kernel_size. Default: None.

    Returns:
        Tensor, output tensor.
    """
    def __init__(self,
                 in_planes,
                 out_planes,
                 kernel_size=4,
                 stride=2,
                 alpha=0.2,
                 norm_mode='batch',
                 pad_mode='CONSTANT',
                 use_relu=True,
                 padding=None):
        super(ConvTransposeNormReLU, self).__init__()
        conv = layers.Conv2dTranspose(in_planes, out_planes, kernel_size, stride=stride, pad_mode='same')
        norm = layers.BatchNorm2d(out_planes)
        if norm_mode == 'instance':
            # Use BatchNorm2d with batchsize=1, affine=False, training=True instead of InstanceNorm2d
            norm = layers.BatchNorm2d(out_planes, affine=False)
        has_bias = (norm_mode == 'instance')
        if padding is None:
            padding = (kernel_size - 1) // 2
        if pad_mode == 'CONSTANT':
            conv = layers.Conv2dTranspose(in_planes, out_planes, kernel_size, stride, pad_mode='same', has_bias=has_bias)
            layer_list = [conv, norm]
        else:
            paddings = ((0, 0), (0, 0), (padding, padding), (padding, padding))
            pad = layers.Pad(paddings=paddings, mode=pad_mode)
            conv = layers.Conv2dTranspose(in_planes, out_planes, kernel_size, stride, pad_mode='pad', has_bias=has_bias)
            layer_list = [pad, conv, norm]
        if use_relu:
            relu = layers.ReLU()
            if alpha > 0:
                relu = layers.LeakyReLU(alpha)
            layer_list.append(relu)
        self.features = layers.SequentialLayer(layer_list)

    def construct(self, x):
        output = self.features(x)
        return output



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""URL check list"""
import numbers
import socket
import ssl
import yaml
from urllib.parse import urlparse
from urllib.request import Request, urlopen, HTTPError

_URL_CHECK_LIST = ('https://tinyms-hub.obs.cn-north-4.myhuaweicloud.com',
                   'https://github.com/tinyms-ai/tinyms')
_URL_CHECK_LIST_INFO = None


class UrlInfo:
    def __init__(self, url):
        parsed = urlparse(url)
        self.domain = parsed.netloc
        self.ip = socket.gethostbyname(parsed.netloc)
        self.path = parsed.path


def _get_checklist_info():
    global _URL_CHECK_LIST_INFO
    if _URL_CHECK_LIST_INFO is None:
        _URL_CHECK_LIST_INFO = set()
        for url in _URL_CHECK_LIST:
            info = UrlInfo(url)
            _URL_CHECK_LIST_INFO.add(info)
    return _URL_CHECK_LIST_INFO


def verify_url(url):
    """
    Verify that the URL is in the url checklist.

    Args:
        url (str): A string of url.

    Returns:
        bool, whether the url in url checklist.
    """
    target_info = UrlInfo(url)
    checklist_info = _get_checklist_info()
    for info in checklist_info:
        if (target_info.ip == info.ip or target_info.domain == info.domain) and \
                target_info.path.startswith(info.path):
            return True
    return False


class HubAssetInfo:
    """
    Information of hub asset model.
    """

    def __init__(self, asset_path):
        asset_dict = ValidHubAsset(asset_path).validate_asset()
        self.name = asset_dict.get('model-name')
        self.backbone_name = asset_dict.get('backbone-name')
        self.type = asset_dict.get('module-type')
        self.fine_tunable = asset_dict.get('fine-tunable')
        self.input_shape = asset_dict.get('input-shape')
        self.author = asset_dict.get('author')
        self.update_time = asset_dict.get('update-time')
        self.repo_link = asset_dict.get('repo-link')
        self.user_id = asset_dict.get('user-id')
        self.backend = asset_dict.get('infer-backend')
        self.dataset = asset_dict.get('train-dataset')
        self.license = asset_dict.get('license')
        self.accuracy = asset_dict.get('accuracy')
        self.used_for = asset_dict.get('used-for')
        self.model_version = asset_dict.get('model-version')
        self.tinyms_version = asset_dict.get('tinyms-version')
        self.asset = asset_dict.get('asset')


class ValidHubAsset:
    r"""
    Check Hub Asset files and extract info.
    """

    def __init__(self, asset_file):
        self.asset_file = asset_file
        self.required_user_fields = ['backbone-name', 'module-type', 'fine-tunable', 'input-shape', 'model-version',
                                     'train-dataset', 'author', 'update-time', 'user-id', 'used-for', 'infer-backend',
                                     'tinyms-version', 'license', 'summary']
        self.optional_backend_field = 'train-backend'
        self.optional_accuracy_field = 'accuracy'

        self.valid_module_type = ['audio', 'cv', 'nlp', 'recommend', 'other']
        self.valid_train_dataset = ['mnist', 'cifar10', 'cifar100', 'mushroom', 'voc2007']
        self.valid_file_format = ['ckpt', 'mindir', 'onnx']
        self.valid_used_for = ['inference', 'transfer-learning']
        self.valid_backend = ['cpu', 'gpu']

        # The current sections list below will be required.
        # If necessary, please add it, like 'Model Description'.
        self.required_sections = []

    def _validate_asset_link_field(self, link):
        r"""
        Make sure the github or gitee repo exists
        """
        if link is None:
            return
        link = link.strip('<>')
        if not verify_url(link):
            raise ValueError('url: ``{}`` is not trust in {}'.format(link, self.asset_file))

        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE
        try:
            req = Request(link)
            urlopen(req, context=ctx)
        except HTTPError:
            raise ValueError('``{}`` is not valid url in {}'.format(link, self.asset_file))

    def _validate_file_format_field(self, file_format):
        r"""
        Only allow file_format in predefined set
        """
        if file_format.lower() not in self.valid_file_format:
            raise ValueError('file_format ``{}`` is not valid in {}. Choose from {}'
                             .format(file_format, self.asset_file, self.valid_file_format))

    def _validate_asset_field(self, asset):
        require_keys = ["asset-link", "asset-sha256", "file-format"]
        for k in require_keys:
            if k not in asset:
                raise ValueError('field: ``{}`` is required in {}, but not found.'
                                 .format(k, self.asset_file))
        self._validate_asset_link_field(asset['asset-link'])
        self._validate_file_format_field(asset['file-format'])

    def _validate_train_dataset_field(self, train_dataset):
        r"""
        Only allow train_dataset in predefined set
        """
        if train_dataset not in self.valid_train_dataset:
            raise ValueError('train_dataset ``{}`` is not valid in {}. Choose from {}'
                             .format(train_dataset, self.asset_file, self.valid_train_dataset))

    def _validate_used_for_field(self, used_for):
        r"""
        Only allow used_for in predefined set
        """
        used_for_list = used_for.split('/')
        for item in used_for_list:
            if item.lower() not in self.valid_used_for:
                raise ValueError('used_for ``{}`` is not valid in {}. Choose from {}'
                                 .format(item, self.asset_file, self.valid_used_for))

    def _validate_backend_field(self, backend):
        r"""
        Only allow backend in predefined set
        """
        if not isinstance(backend, list):
            backend = [backend]
        for bk in backend:
            if bk.lower() not in self.valid_backend:
                raise ValueError('backend ``{}`` is not valid in {}. Choose from {}'
                                 .format(bk, self.asset_file, self.valid_backend))

    def _validate_module_type_field(self, module_type):
        r"""
        Only allow module_type in predefined set
        """
        items = module_type.lower().split('-')
        if len(items) > 2:
            raise Exception("module-type could only no more than one '-' ")
        first_class = items[0]
        if first_class not in self.valid_module_type:
            raise ValueError('module_type ``{}`` is not valid in {}. Valid module_type set is {}'
                             .format(module_type, self.asset_file, self.valid_module_type))

    def _validate_header(self, header):
        r"""
        Make sure the header is in the required format
        """
        for field in self.required_user_fields:
            if field not in header:
                raise ValueError("field: ``{}`` is required in {}, but not found."
                                 .format(field, self.asset_file))

        if not isinstance(header['fine-tunable'], bool):
            raise TypeError("`fine-tunable` must be `bool`, but got {}".format(header['fine-tunable']))

        if not isinstance(header['input-shape'], list):
            raise TypeError("`input-shape` must be `list` of `int`, but got {}".format(header['input-shape']))
        for i in header['input-shape']:
            if not isinstance(i, int) and not isinstance(i, list):
                raise TypeError("`input-shape` must be `list` of `int`, but got {}".format(header['input-shape']))

        if not isinstance(header['infer-backend'], list):
            raise TypeError("`infer-backend` must be `list` of `str`, but got {}".format(header['infer-backend']))
        for i in header['infer-backend']:
            if not isinstance(i, str) and not isinstance(i, list):
                raise TypeError("`infer-backend` must be `list` of `str`, but got {}".format(header['infer-backend']))

        self._validate_train_dataset_field(header['train-dataset'])
        self._validate_used_for_field(header['used-for'])
        self._validate_backend_field(header['infer-backend'])
        self._validate_module_type_field(header['module-type'])
        if not header.get('asset') is None:
            self._validate_asset_field(header['asset'])

        if self.optional_accuracy_field in header.keys():
            if not isinstance(header[self.optional_accuracy_field], numbers.Number):
                raise TypeError("`accuracy` must be `number`, but got {}".
                                format(header[self.optional_accuracy_field]))
        if self.optional_backend_field in header.keys():
            self._validate_backend_field(header[self.optional_backend_field])

    def validate_asset(self):
        with open(self.asset_file) as f:
            asset_dict = yaml.load(f, Loader=yaml.FullLoader)
        if not asset_dict:
            raise TypeError("Failed to parse a valid yaml header")

        self._validate_header(asset_dict)
        return asset_dict



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import matplotlib.pyplot as plt
import numpy as np
from PIL import Image


class ImageViewer():
    r'''
    ImageViewer is a class defined for visualizing the input image.

    Args:
        image (Union[PIL.Image, numpy.ndarray]): image input.
        label (str, optional): specifies the label of this image. Default: None.

    Raises:
        TypeError: When image input is not `numpy.ndarray` or `PIL.Image`.
    '''

    def __init__(self, image, label=None):
        if not isinstance(image, (np.ndarray, Image.Image)):
            raise TypeError("Input type should be numpy.ndarray or PIL.Image, got {}.".format(type(image)))
        if isinstance(image, Image.Image):
            image = np.array(image)
        self._image = image
        self._label = label

    @property
    def image(self):
        return self._image

    @property
    def label(self):
        return self._label

    def show(self):
        r'''
        Directly show the visualized picture.

        Note:
            This function is only valid when being called in interactive environment, such like `Jupyter notebook`.

        Examples:
            >>> form PIL import Image
            >>>
            >>> img = Image.open('example.jpg')
            >>> img_viewer = ImageViewer(img, label='cat')
            >>> img_viewer.show()
        '''
        plt.figure(figsize=(16, 10))
        if self._image.ndim == 2:
            plt.imshow(self._image, cmap='gray')
        else:
            plt.imshow(self._image)

        if not self._label is None:
            plt.title("label: %s" % self._label)
            plt.show()

    def draw(self, pred_res, labels):
        r'''
        Draw the predicting boxes on the picture and show the visualized picture.

        Args:
            pred_res (dict): The predcition result from `tinyms.serving.predict` method.
            labels (list): The labels should be input manually with a list of string.
                This argument is required for distinguishing the color from different classes.

        Note:
            This function is only valid when being called in interactive environment, such like `Jupyter notebook`.

        Examples:
            >>> form PIL import Image
            >>>
            >>> img = Image.open('example.jpg')
            >>> img_viewer = ImageViewer(img)
            >>> labels = ['1', '2', '3']
            >>> img_viewer.draw(pred_res, labels)
        '''
        colors = plt.cm.hsv(np.linspace(0, 1, len(labels)+1)).tolist()
        plt.figure(figsize=(20, 12))
        plt.imshow(np.squeeze(self._image))

        current_axis = plt.gca()

        for sample in pred_res:
            xmin = sample['bbox'][0]
            ymin = sample['bbox'][1]
            width = sample['bbox'][2]
            height = sample['bbox'][3]
            category_id = sample['category_id']
            cls = labels.index(category_id)
            color = colors[cls]

            label = '{}: {:.2f}'.format(category_id, sample['score'])
            current_axis.add_patch(plt.Rectangle((xmin, ymin), width, height,
                                                 color=color, fill=False, linewidth=2))
            current_axis.text(xmin, ymin, label, size='x-large', color='white',
                              bbox={'facecolor': color, 'alpha': 1.0})



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
import tinyms as ts

from . import _transform_ops
from ._transform_ops import *
from ..data import BertDataset


__all__ = [
    'bert_transform', 'BertDatasetTransform',
]
__all__.extend(_transform_ops.__all__)


class BertDatasetTransform(object):
    r'''
    Apply preprocess operation on GeneratorDataset instance.
    '''
    def __init__(self):
        pass

    def apply_ds(self, data_set, batch_size):

        assert isinstance(data_set, BertDataset), "For BertDatasetTransform, BertDataset is needed"

        type_cast_op = TypeCast(ts.int32)
        data_set = data_set.map(operations=type_cast_op, input_columns="masked_lm_ids")
        data_set = data_set.map(operations=type_cast_op, input_columns="masked_lm_positions")
        data_set = data_set.map(operations=type_cast_op, input_columns="next_sentence_labels")
        data_set = data_set.map(operations=type_cast_op, input_columns="segment_ids")
        data_set = data_set.map(operations=type_cast_op, input_columns="input_mask")
        data_set = data_set.map(operations=type_cast_op, input_columns="input_ids")
        # apply batch operations
        data_set = data_set.batch(batch_size, drop_remainder=True)

        return data_set


bert_transform = BertDatasetTransform()


# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The client for TinyMS serving """
import os
import json
import sys
import socket
import requests
import numpy as np
from PIL import Image
from tinyms.vision import mnist_transform, cifar10_transform, imagefolder_transform, voc_transform, cyclegan_transform
from tinyms.data.utils import load_resized_img

transform_checker = {
    'mnist': mnist_transform,
    'cifar10': cifar10_transform,
    'imagenet2012': imagefolder_transform,
    'voc': voc_transform,
    'cityscape': cyclegan_transform,
}


class Client:
    '''
    Client is the entrance to connecting to serving server, and also
    send all requests to server side by HTTP request.

    Args:
        host (str): Serving server host ip. Default: '127.0.0.1'.
        port (int): Serving server listen port. Default: 5000.

    Examples:
        >>> from tinyms.serving import Client
        >>>
        >>> client = Client()
    '''

    def __init__(self, host='127.0.0.1', port=5000):
        self.host = host
        self.port = port

    def _server_started(self):
        """
        Detect whether the serving server is started or not.

        A bool value of True will be returned if the server is started, else False.

        Returns:
            A bool value of True(if server started) or False(if server not started).
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((self.host, self.port))
            s.shutdown(2)
            return True
        except:
            return False

    def list_servables(self):
        """
        List the model that is currently served by the backend server.

        A `GET` request will be sent to the server (127.0.0.1:5000) which will then
        be routed to 127.0.0.1:5000/servables, and the backend servable information
        will be returned to the client.

        Returns:
            res_body['servables'] (str) will be returned, the backend servable information.
            Error message will be returned and printed if requests.status_code is not ok.
            'Server not started' will be returned if server is not started

        Examples:
            >>> # Running the quickstart tutorial, after server started and servable json defined
            >>> from tinyms.serving import Client
            >>>
            >>> client = Client()
            >>> client.list_servables()
            [{
                'description': 'This servable hosts a lenet5 model predicting numbers',
                'model': {
                    'class_num': 10,
                    'format': 'ckpt',
                    'name': 'lenet5'
                },
                'name': 'lenet5'
            }]
        """
        if not self._server_started() is True:
            print('Server not started at host %s, port %d' % (self.host, self.port))
            sys.exit(0)
        else:
            headers = {'Content-Type': 'application/json'}
            url = 'http://'+self.host+':'+str(self.port)+'/servables'
            res = requests.get(url=url, headers=headers)
            res.content.decode("utf-8")
            res_body = res.json()

            if res.status_code != requests.codes.ok:
                print("Request error! Status code: ", res.status_code)
                sys.exit(0)
            elif res_body['status'] != 0:
                print(res_body['err_msg'])
                sys.exit(0)
            else:
                return res_body['servables']

    def predict(self, img_path, servable_name, dataset_name="mnist", strategy="TOP1_CLASS"):
        """
        Send the predict request to the backend server, get the return value and do the post process

        Predict the input image, and get the result. User must specify the image_path, servable_name, dataset_name and output_strategy to get the predict result.

        Args:
            img_path (str): path to the image
            servable_name (str): the `name` in `servable_json`, now supports 6 servables: `lenet5`, `resnet50_imagenet2012`, `resnet50_cifar10`, `mobilenetv2`, `ssd300` and `cyclegan_cityscape`.
            dataset_name (str): the name of the dataset that is used to train the model, now supports 5 datasets: `mnist`, `imagenet2012`, `cifar10`, `voc`, `cityscape`
            strategy (str): the output strategy, for lenet5, resnet50 and mobilenetv2, select between 'TOP1_CLASS' and 'TOP5_CLASS', for ssd300, only `TOP1_CLASS`, for cyclegan_cityscape, select between `gray2color` and `color2gray`

        Returns:
            For lenet5, resnet50, mobilenetv2, the output is a string of predict result.
            For ssd300, the output is a string of bounding boxes coordinates and labels, which can be further processed using `ImageViewer` function
            For cyclegan, the output is a numpy of image, which can be transformed to image using `Image.fromarray`

        Examples:
            >>> # Running the quickstart tutorial, after server started and servable json defined
            >>> from tinyms.serving import Client
            >>>
            >>> client = Client()
            >>> print(client.predict('/root/7.png', 'lenet5', 'mnist', 'TOP1_CLASS'))
            TOP1: 7, score: 0.99943381547927856445
        """
        # Check if args are valid
        if not os.path.isfile(img_path):
            print("The image path {} not exist!".format(img_path))
            sys.exit(0)
        trans_func = transform_checker.get(dataset_name)
        if trans_func is None:
            print("Currently dataset_name only supports {}!".format(list(transform_checker.keys())))
            sys.exit(0)
        if strategy not in ("TOP1_CLASS", "TOP5_CLASS", "gray2color", "color2gray"):
            print("Currently strategy only supports `TOP1_CLASS`, `TOP5_CLASS`, `gray2color` and`color2gray`!")
            sys.exit(0)

        # Perform the transform operation for the input image
        if servable_name == 'cyclegan_cityscape':
            img = np.array(load_resized_img(img_path))
        else:
            img = Image.open(img_path)
        img_data = trans_func(img)

        if not self._server_started() is True:
            print('Server not started at host %s, port %d' % (self.host, self.port))
            sys.exit(0)
        else:
            # Construct the request payload
            payload = {
                'instance': {
                    'shape': list(img_data.shape),
                    'dtype': img_data.dtype.name,
                    'data': json.dumps(img_data.tolist())
                },
                'strategy': strategy
            }
            headers = {'Content-Type': 'application/json'}
            url = 'http://'+self.host+':'+str(self.port)+'/servables/' + servable_name
            res = requests.post(url=url, headers=headers, data=json.dumps(payload))
            res.content.decode("utf-8")
            res_body = res.json()

            if res.status_code != requests.codes.ok:
                print("Request error! Status code: ", res.status_code)
                sys.exit(0)
            elif res_body['status'] != 0:
                print(res_body['err_msg'])
                sys.exit(0)
            else:
                instance = res_body['instance']
                res_data = np.array(json.loads(instance['data']))
                if dataset_name == 'voc':
                    iw, ih = img.size
                    data = trans_func.postprocess(res_data, (ih, iw), strategy)
                elif dataset_name == 'cityscape':
                    data = res_data
                else:
                    data = trans_func.postprocess(res_data, strategy)
                return data



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""UNet Generator."""

from tinyms import layers
from tinyms.primitives import Concat


class UnetGenerator(layers.Layer):
    """
    Unet-based generator.

    Args:
        in_planes (int): the number of channels in input images.
        out_planes (int): the number of channels in output images.
        ngf (int): the number of filters in the last conv layer.
        n_layers (int): the number of downsamplings in UNet.
        alpha (float): LeakyRelu slope. Default: 0.2.
        norm_mode (str): Specifies norm method. The optional values are "batch", "instance".
        dropout (bool): Use dropout or not. Default: False.

    Returns:
        Tensor, output tensor.
    """
    def __init__(self, in_planes, out_planes, ngf=64, n_layers=7, alpha=0.2, norm_mode='bn', dropout=False):
        super(UnetGenerator, self).__init__()
        # construct unet structure
        unet_block = UnetSkipConnectionBlock(ngf * 8, ngf * 8, in_planes=None, submodule=None,
                                             norm_mode=norm_mode, innermost=True)
        for _ in range(n_layers - 5):
            unet_block = UnetSkipConnectionBlock(ngf * 8, ngf * 8, in_planes=None, submodule=unet_block,
                                                 norm_mode=norm_mode, dropout=dropout)
        # gradually reduce the number of filters from ngf * 8 to ngf
        unet_block = UnetSkipConnectionBlock(ngf * 4, ngf * 8, in_planes=None, submodule=unet_block,
                                             norm_mode=norm_mode)
        unet_block = UnetSkipConnectionBlock(ngf * 2, ngf * 4, in_planes=None, submodule=unet_block,
                                             norm_mode=norm_mode)
        unet_block = UnetSkipConnectionBlock(ngf, ngf * 2, in_planes=None, submodule=unet_block, norm_mode=norm_mode)
        self.model = UnetSkipConnectionBlock(out_planes, ngf, in_planes=in_planes, submodule=unet_block,
                                             outermost=True, norm_mode=norm_mode)

    def construct(self, x):
        return self.model(x)


class UnetSkipConnectionBlock(layers.Layer):
    """Unet submodule with skip connection.

    Args:
        outer_nc (int): The number of filters in the outer conv layer
        inner_nc (int): The number of filters in the inner conv layer
        in_planes (int): The number of channels in input images/features
        dropout (bool): Use dropout or not. Default: False.
        submodule (Layer): Previously defined submodules
        outermost (bool): If this module is the outermost module
        innermost (bool): If this module is the innermost module
        alpha (float): LeakyRelu slope. Default: 0.2.
        norm_mode (str): Specifies norm method. The optional values are "batch", "instance".

    Returns:
        Tensor, output tensor.
    """
    def __init__(self, outer_nc, inner_nc, in_planes=None, dropout=False,
                 submodule=None, outermost=False, innermost=False, alpha=0.2, norm_mode='batch'):
        super(UnetSkipConnectionBlock, self).__init__()
        downnorm = layers.BatchNorm2d(inner_nc)
        upnorm = layers.BatchNorm2d(outer_nc)
        use_bias = False
        if norm_mode == 'instance':
            downnorm = layers.BatchNorm2d(inner_nc, affine=False)
            upnorm = layers.BatchNorm2d(outer_nc, affine=False)
            use_bias = True
        if in_planes is None:
            in_planes = outer_nc
        downconv = layers.Conv2d(in_planes, inner_nc, kernel_size=4,
                             stride=2, padding=1, has_bias=use_bias, pad_mode='pad')
        downrelu = layers.LeakyReLU(alpha)
        uprelu = layers.ReLU()

        if outermost:
            upconv = layers.Conv2dTranspose(inner_nc * 2, outer_nc,
                                            kernel_size=4, stride=2,
                                            padding=1, pad_mode='pad')
            down = [downconv]
            up = [uprelu, upconv, layers.Tanh()]
            model = down + [submodule] + up
        elif innermost:
            upconv = layers.Conv2dTranspose(inner_nc, outer_nc,
                                            kernel_size=4, stride=2,
                                            padding=1, has_bias=use_bias, pad_mode='pad')
            down = [downrelu, downconv]
            up = [uprelu, upconv, upnorm]
            model = down + up
        else:
            upconv = layers.Conv2dTranspose(inner_nc * 2, outer_nc,
                                            kernel_size=4, stride=2,
                                            padding=1, has_bias=use_bias, pad_mode='pad')
            down = [downrelu, downconv, downnorm]
            up = [uprelu, upconv, upnorm]

            model = down + [submodule] + up
            if dropout:
                model.append(layers.Dropout(0.5))

        self.model = layers.SequentialLayer(model)
        self.skip_connections = not outermost
        self.concat = Concat(axis=1)

    def construct(self, x):
        out = self.model(x)
        if self.skip_connections:
            out = self.concat((out, x))
        return out



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
import numpy as np

from tinyms import layers, Tensor
from tinyms import Parameter

from tinyms.common import dtype
from tinyms.layers import Dropout
from tinyms.primitives import BiasAdd, Cast, MatMul, ReLU, Sigmoid, Tanh
from tinyms.primitives import Concat, Gather, Mul, ReduceSum, Reshape, Square, Tile


def _init_activation(activation):
    activation = str(activation).lower()
    if activation == "relu":
        activation_func = ReLU()
    elif activation == "sigmoid":
        activation_func = Sigmoid()
    elif activation == "tanh":
        activation_func = Tanh()
    else:
        raise ValueError("activation type: {} not supported!".format(activation))

    return activation_func


class DenseLayer(layers.Layer):
    """
    DeepFM denselayer with convert dtypes definition.

    Args:
        in_channels (int): Input channels.
        out_channels (int): Output channels.
        keep_prob (float): Keep prob. Default: 1.0.
        convert_dtype (bool): Whether to convert data type. Defalut: True.

    Returns:
        Tensor, output tensor.

    Examples:
        >>> DenseLayer(3, 256)
    """
    def __init__(self, in_channels, out_channels, keep_prob=0.9, convert_dtype=True):
        super(DenseLayer, self).__init__()
        self.weight = Parameter(
            Tensor(np.random.normal(loc=0.0, scale=0.01, size=[in_channels, out_channels]).astype(dtype=np.float32)),
            name="weight")
        self.bias = Parameter(
            Tensor(np.random.normal(loc=0.0, scale=0.01, size=[out_channels]).astype(dtype=np.float32)),
            name="bias")
        self.convert_dtype = convert_dtype
        self.dropout = Dropout(keep_prob=keep_prob)
        self.cast = Cast()
        self.matmul = MatMul(transpose_b=False)
        self.bias_add = BiasAdd()
        self.activation = ReLU()

    def construct(self, x):
        """
        Construct function
        """
        x = self.dropout(x)
        if self.convert_dtype:
            x = self.cast(x, dtype.float16)
            weight = self.cast(self.weight, dtype.float16)
            bias = self.cast(self.bias, dtype.float16)
            wx = self.matmul(x, weight)
            wx = self.bias_add(wx, bias)
            if self.activation is not None:
                wx = self.activation(wx)
            wx = self.cast(wx, dtype.float32)
        else:
            wx = self.matmul(x, self.weight)
            wx = self.bias_add(wx, self.bias)
            if self.activation:
                wx = self.activation(wx)
        return wx


class DeepFM(layers.Layer):
    """
    DeepFM architecture.

    Args:
        field_size (int): The field size.
        vocab_size (int): The vocabulary size.
        embed_size (int): The embeding size.
        keep_prob (float): The keep prob value. Default: 0.9.
        convert_dtype (bool): Whether to convert data type.
        CPU can only set to False. Defalut: False.

    Returns:
        Tensor, output tensor.

    Examples:
        >>> from tinyms.model import DeepFM
        >>>
        >>> DeepFM(field_size=39, vocab_size=184965, embed_size=80)

    """
    def __init__(self, field_size, vocab_size, embed_size, keep_prob=0.9, convert_dtype=False):
        super(DeepFM, self).__init__()
        self.field_size = field_size
        self.vocab_size = vocab_size
        self.embed_size = embed_size
        self.embedding = Parameter(
            Tensor(np.random.normal(loc=0.0, scale=0.01, size=[vocab_size, embed_size]).astype(dtype=np.float32)),
            name="embedding")
        self.fm_weight = Parameter(
            Tensor(np.random.normal(loc=0.0, scale=0.01, size=[vocab_size, 1]).astype(dtype=np.float32)),
            name="fm_weight")
        self.dense_layer_1 = DenseLayer(field_size * embed_size, 1024, keep_prob=keep_prob, convert_dtype=convert_dtype)
        self.dense_layer_2 = DenseLayer(1024, 512, keep_prob=keep_prob, convert_dtype=convert_dtype)
        self.dense_layer_3 = DenseLayer(512, 256, keep_prob=keep_prob, convert_dtype=convert_dtype)
        self.dense_layer_4 = DenseLayer(256, 128, keep_prob=keep_prob, convert_dtype=convert_dtype)
        self.dense_layer_5 = DenseLayer(128, 1, keep_prob=keep_prob, convert_dtype=convert_dtype)
        self.gather = Gather()
        self.mul = Mul()
        self.reduce_sum = ReduceSum(keep_dims=False)
        self.reshape = Reshape()
        self.square = Square()
        self.tile = Tile()
        self.concat = Concat(axis=1)
        self.cast = Cast()

    def construct(self, ids, wts):
        mask = self.reshape(wts, (-1, self.field_size, 1))
        # Linear layer
        fm_id_weight = self.gather(self.fm_weight, ids, 0)
        wx = self.mul(fm_id_weight, mask)
        linear_out = self.reduce_sum(wx, 1)
        # FM layer
        fm_id_embeds = self.gather(self.embedding, ids, 0)
        vx = self.mul(fm_id_embeds, mask)
        v1 = self.reduce_sum(vx, 1)
        v1 = self.square(v1)
        v2 = self.square(vx)
        v2 = self.reduce_sum(v2, 1)
        fm_out = 0.5 * self.reduce_sum(v1 - v2, 1)
        fm_out = self.reshape(fm_out, (-1, 1))
        # Deep layer
        deep_in = self.reshape(vx, (-1, self.field_size * self.embed_size))
        deep_in = self.dense_layer_1(deep_in)
        deep_in = self.dense_layer_2(deep_in)
        deep_in = self.dense_layer_3(deep_in)
        deep_in = self.dense_layer_4(deep_in)
        deep_out = self.dense_layer_5(deep_in)
        out = linear_out + fm_out + deep_out
        return out, self.fm_weight, self.embedding


def deepfm(**kwargs):
    """
    Get DeepFM neural network.

    Args:
        field_size (int): The field size. Default: 39.
        vocab_size (int): The vocabulary size. Default: 184965.
        embed_size (int): The embeding size. Default: 80.
        keep_prob (float): The keep prob value. Default: 0.9.
        convert_dtype (bool): Whether to convert data type. Defalut: False.

    Returns:
        layers.Layer, layer instance of DeepFM neural network.

    Examples:
        >>> from tinyms.model import deepfm
        >>>
        >>> net = deepfm(39, 184965, 80)
    """
    return DeepFM(field_size=kwargs.get('field_size', 39),
                  vocab_size=kwargs.get('vocab_size', 184965),
                  embed_size=kwargs.get('embed_size', 80),
                  keep_prob=kwargs.get('keep_prob', 0.9),
                  convert_dtype=kwargs.get('convert_dtype', False))



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from .. import layers
from ..initializers import Normal


class LeNet(layers.Layer):
    """
    LeNet architecture.

    Args:
        class_num (int): The number of classes that the training images are belonging to.
        channel_num (int): The channel number.
    Returns:
        Tensor, output tensor.

    Examples:
        >>> LeNet(class_num=10)
    """

    def __init__(self, class_num=10, channel_num=1):
        super(LeNet, self).__init__()
        self.conv1 = layers.Conv2d(channel_num, 6, 5, pad_mode='valid')
        self.conv2 = layers.Conv2d(6, 16, 5, pad_mode='valid')
        self.relu = layers.ReLU()
        self.max_pool2d = layers.MaxPool2d(kernel_size=2, stride=2)
        self.flatten = layers.Flatten()
        self.fc1 = layers.Dense(16 * 5 * 5, 120, weight_init=Normal(0.02))
        self.fc2 = layers.Dense(120, 84, weight_init=Normal(0.02))
        self.fc3 = layers.Dense(84, class_num, weight_init=Normal(0.02))

    def construct(self, x):
        x = self.max_pool2d(self.relu(self.conv1(x)))
        x = self.max_pool2d(self.relu(self.conv2(x)))
        x = self.flatten(x)
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.fc3(x)
        return x


def lenet5(**kwargs):
    """
    Get LeNet5 neural network.

    Args:
        class_num (int): Class number. Default: 10.

    Returns:
        layers.Layer, layer instance of LeNet5 neural network.

    Examples:
        >>> net = lenet5(class_num=10)
    """
    return LeNet(class_num=kwargs.get('class_num', 10))



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from scipy.stats import truncnorm

import tinyms as ts
from tinyms import layers, Tensor
from tinyms.primitives import tensor_add, ReduceMean


def _conv_variance_scaling_initializer(in_channel, out_channel, kernel_size):
    fan_in = in_channel * kernel_size * kernel_size
    scale = 1.0
    scale /= max(1., fan_in)
    stddev = (scale ** 0.5) / .87962566103423978
    mu, sigma = 0, stddev
    weight = truncnorm(-2, 2, loc=mu, scale=sigma).rvs(out_channel * in_channel * kernel_size * kernel_size)
    return ts.reshape(weight, (out_channel, in_channel, kernel_size, kernel_size))


def _weight_variable(shape, factor=0.01):
    init_value = np.random.randn(*shape).astype(np.float32) * factor
    return Tensor(init_value)


def _conv3x3(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 3, 3)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=3, stride=stride, padding=0, pad_mode='same', weight_init=weight)


def _conv1x1(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 1, 1)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=1, stride=stride, padding=0, pad_mode='same', weight_init=weight)


def _conv7x7(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 7, 7)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=7, stride=stride, padding=0, pad_mode='same', weight_init=weight)


def _bn(channel):
    return layers.BatchNorm2d(channel, eps=1e-4, momentum=0.9,
                              gamma_init=1, beta_init=0, moving_mean_init=0, moving_var_init=1)


def _bn_last(channel):
    return layers.BatchNorm2d(channel, eps=1e-4, momentum=0.9,
                              gamma_init=0, beta_init=0, moving_mean_init=0, moving_var_init=1)


def _fc(in_channel, out_channel):
    weight_shape = (out_channel, in_channel)
    weight = _weight_variable(weight_shape)
    return layers.Dense(in_channel, out_channel, has_bias=True, weight_init=weight, bias_init=0)


class ResidualBlock(layers.Layer):
    """
    ResNet V1 residual block definition.

    Args:
        in_channel (int): Input channel.
        out_channel (int): Output channel.
        stride (int): Stride size for the first convolutional layer. Default: 1.

    Returns:
        Tensor, output tensor.

    Examples:
        >>> ResidualBlock(3, 256, stride=2)
    """
    expansion = 4

    def __init__(self, in_channel, out_channel, stride=1):
        super(ResidualBlock, self).__init__()
        channel = out_channel // self.expansion
        self.conv1 = _conv1x1(in_channel, channel, stride=1)
        self.bn1 = _bn(channel)
        self.conv2 = _conv3x3(channel, channel, stride=stride)
        self.bn2 = _bn(channel)
        self.conv3 = _conv1x1(channel, out_channel, stride=1)
        self.bn3 = _bn_last(out_channel)
        self.relu = layers.ReLU()

        self.down_sample = False
        self.down_sample_layer = None
        if stride != 1 or in_channel != out_channel:
            self.down_sample = True
        if self.down_sample:
            self.down_sample_layer = layers.SequentialLayer(
                [_conv1x1(in_channel, out_channel, stride), _bn(out_channel)])

    def construct(self, x):
        out = self.relu(self.bn1(self.conv1(x)))
        out = self.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        identity = x
        if self.down_sample:
            identity = self.down_sample_layer(identity)
        out = self.relu(tensor_add(out, identity))

        return out


class ResNet(layers.Layer):
    """
    ResNet architecture.

    Args:
        block (layers.Layer): Block for network.
        layer_nums (list): Numbers of block in different layers.
        in_channels (list): Input channel in each layer.
        out_channels (list): Output channel in each layer.
        strides (list):  Stride size in each layer.
        num_classes (int): The number of classes that the training images are belonging to.

    Returns:
        Tensor, output tensor.

    Examples:
        >>> ResNet(ResidualBlock,
        >>>        [3, 4, 6, 3],
        >>>        [64, 256, 512, 1024],
        >>>        [256, 512, 1024, 2048],
        >>>        [1, 2, 2, 2],
        >>>        10)
    """

    def __init__(self,
                 block,
                 layer_nums,
                 in_channels,
                 out_channels,
                 strides,
                 num_classes):
        super(ResNet, self).__init__()

        if not len(layer_nums) == len(in_channels) == len(out_channels) == 4:
            raise ValueError("the length of layer_num, in_channels, out_channels list must be 4!")

        self.conv1 = _conv7x7(3, 64, stride=2)
        self.bn1 = _bn(64)
        self.relu = layers.ReLU()
        self.maxpool = layers.MaxPool2d(kernel_size=3, stride=2, pad_mode="same")
        self.layer1 = self._make_layer(block,
                                       layer_nums[0],
                                       in_channel=in_channels[0],
                                       out_channel=out_channels[0],
                                       stride=strides[0])
        self.layer2 = self._make_layer(block,
                                       layer_nums[1],
                                       in_channel=in_channels[1],
                                       out_channel=out_channels[1],
                                       stride=strides[1])
        self.layer3 = self._make_layer(block,
                                       layer_nums[2],
                                       in_channel=in_channels[2],
                                       out_channel=out_channels[2],
                                       stride=strides[2])
        self.layer4 = self._make_layer(block,
                                       layer_nums[3],
                                       in_channel=in_channels[3],
                                       out_channel=out_channels[3],
                                       stride=strides[3])

        self.mean = ReduceMean(keep_dims=True)
        self.flatten = layers.Flatten()
        self.end_point = _fc(out_channels[3], num_classes)

    def _make_layer(self, block, layer_num, in_channel, out_channel, stride):
        """
        Make stage network of ResNet.

        Args:
            block (layers.Layer): Resnet block.
            layer_num (int): Layer number.
            in_channel (int): Input channel.
            out_channel (int): Output channel.
            stride (int): Stride size for the first convolutional layer.

        Returns:
            SequentialLayer, the output layer.

        Examples:
            >>> _make_layer(ResidualBlock, 3, 128, 256, 2)
        """
        layer = layers.SequentialLayer([block(in_channel, out_channel, stride=stride)])
        for _ in range(1, layer_num):
            resnet_block = block(out_channel, out_channel, stride=1)
            layer.append(resnet_block)

        return layer

    def construct(self, x):
        c1 = self.maxpool(self.relu(self.bn1(self.conv1(x))))
        c2 = self.layer1(c1)
        c3 = self.layer2(c2)
        c4 = self.layer3(c3)
        c5 = self.layer4(c4)
        out = self.end_point(self.flatten(self.mean(c5, (2, 3))))

        return out


def resnet50(**kwargs):
    """
    Get ResNet50 neural network.

    Args:
        class_num (int): Class number. Default: 10.

    Returns:
        layers.Layer, layer instance of ResNet50 neural network.

    Examples:
        >>> net = resnet50(10)
    """
    return ResNet(ResidualBlock,
                  [3, 4, 6, 3],
                  [64, 256, 512, 1024],
                  [256, 512, 1024, 2048],
                  [1, 2, 2, 2],
                  kwargs.get('class_num', 10))
